"""Tests for lookup handler for var."""
from unittest import TestCase

from runway.lookups.handlers.var import VarLookup
from runway.util import MutableMap

VARIABLES = MutableMap(**{
    'str_val': 'test',
    'false_val': False
})


class TestVarLookup(TestCase):
    """Tests for VarLookup."""

    def test_handle(self):
        """Validate handle base functionality."""
        query = 'str_val'
        result = VarLookup.handle(query, context=None, variables=VARIABLES)

        self.assertEqual(result, 'test')

    def test_handle_false_result(self):
        """Validate that a bool value of False can be resolved."""
        query = 'false_val'
        result = VarLookup.handle(query, context=None, variables=VARIABLES)

        self.assertFalse(result)

    def test_handle_not_found(self):
        """Validate exception when lookup cannot be resolved."""
        query = 'NOT_VALID'

        with self.assertRaises(ValueError):
            VarLookup.handle(query, context=None, variables=VARIABLES)



"""Test to verify behavior on declining deployment"""
import sys

import pexpect
from runway.util import change_dir

from integration_tests.test_cdk.test_cdk import CDK


class TestDeclineDeploy(CDK):
    """Test to verify behavior on declining deployment"""

    TEST_NAME = __name__

    def deploy(self):
        """Deploy provider."""
        self.copy_fixture('decline-deploy-app.cdk')
        self.copy_runway('decline-deploy')
        with change_dir(self.cdk_test_dir):
            child = pexpect.spawn('runway deploy')
            try:
                child.logfile = sys.stdout.buffer
                child.expect('Do you wish to deploy these changes (y/n)?', timeout=120)
                child.sendline('n')
                i = child.expect('Aborted')
            except pexpect.EOF:
                self.logger.debug('EOF Reached')
                return 0
            else:
                return i

    def run(self):
        """Run tests."""
        self.clean()
        assert self.deploy() == 0, '{}: Declining Deployment failed'.format(__name__)

    def teardown(self):
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        self.clean()



"""Destroys CloudFormation stacks based on the given config.

CFNgin will determine the order in which stacks should be destroyed based on
any manual requirements they specify or output values they rely on from other
stacks.

"""
from ...actions import destroy
from .base import BaseCommand, cancel


class Destroy(BaseCommand):
    """Destroy subcommand."""

    name = "destroy"
    description = __doc__

    def add_arguments(self, parser):
        """Add arguments."""
        super(Destroy, self).add_arguments(parser)
        self._add_argument_max_parallel(parser)
        self._add_argument_tail(parser)
        parser.add_argument("-f", "--force", action="store_true",
                            help="Whether or not you want to go through "
                                 " with destroying the stacks")
        parser.add_argument("--targets", "--stacks", action="append",
                            metavar="STACKNAME", type=str,
                            help="Only work on the stacks given. Can be "
                                 "specified more than once. If not specified "
                                 "then CFNgin will work on all stacks in the "
                                 "config file.")

    def run(self, options):
        """Run the command."""
        super(Destroy, self).run(options)
        action = destroy.Action(options.context,
                                provider_builder=options.provider_builder,
                                cancel=cancel())
        action.execute(concurrency=options.max_parallel,
                       force=options.force,
                       tail=options.tail)

    def get_context_kwargs(self, options):
        """Return a dictionary of kwargs that will be used with the Context.

        This allows commands to pass in any specific arguments they define to
        the context.

        Args:
            options (:class:`argparse.Namespace`): arguments that have been
                passed via the command line

        Returns:
            Dict[str, Any]: Dictionary that will be passed to Context
            initializer as kwargs.

        """
        return {"stack_names": options.targets}



"""Register test handlers."""
from typing import Type, Union  # pylint: disable=unused-import

from six import string_types

from runway.util import load_object_from_string

from .handlers import env, ssm, var
from .handlers.base import LookupHandler  # noqa: F401 pylint: disable=W

RUNWAY_LOOKUP_HANDLERS = {}


def register_lookup_handler(lookup_type, handler_or_path):
    # type: (str, Union[Type[LookupHandler], str]) -> None
    """Register a lookup handler.

    Args:
        lookup_type: Name to register the handler under
        handler_or_path: a function or a path to a handler

    """
    handler = handler_or_path
    if isinstance(handler_or_path, string_types):
        handler = load_object_from_string(handler_or_path)
    RUNWAY_LOOKUP_HANDLERS[lookup_type] = handler


def unregister_lookup_handler(lookup_type):
    # type: (str) -> None
    """Unregister the specified test type.

    This is useful when testing various lookup types if you want to unregister
    the lookup type after the test runs.

    Args:
        lookup_type: Name of the lookup type to unregister

    """
    RUNWAY_LOOKUP_HANDLERS.pop(lookup_type, None)


register_lookup_handler(env.TYPE_NAME, env.EnvLookup)
register_lookup_handler(ssm.TYPE_NAME, ssm.SsmLookup)
register_lookup_handler(var.TYPE_NAME, var.VarLookup)



"""CFNgin UI manipulation."""
import logging
import threading
from getpass import getpass

from six.moves import input

LOGGER = logging.getLogger(__name__)


def get_raw_input(message):
    """Just a wrapper for :func:`input` for testing purposes."""
    return input(message)


class UI(object):
    """Used internally from terminal output in a multithreaded environment.

    Ensures that two threads don't write over each other while asking a user
    for input (e.g. in interactive mode).

    """

    def __init__(self):
        """Instantiate class."""
        self._lock = threading.RLock()

    def lock(self, *_args, **_kwargs):
        """Obtain an exclusive lock on the UI for the current thread."""
        return self._lock.acquire()

    def unlock(self, *_args, **_kwargs):
        """Release the lock on the UI."""
        return self._lock.release()

    def info(self, *args, **kwargs):
        """Log the line if the current thread owns the underlying lock."""
        self.lock()
        try:
            return LOGGER.info(*args, **kwargs)
        finally:
            self.unlock()

    def ask(self, message):
        """Collect input from a user in a multithreaded environment.

        This wraps the built-in raw_input function to ensure that only 1
        thread is asking for input from the user at a give time. Any process
        that tries to log output to the terminal will be blocked while the
        user is being prompted.

        """
        self.lock()
        try:
            return get_raw_input(message)
        finally:
            self.unlock()

    def getpass(self, *args):
        """Wrap getpass to lock the UI."""
        try:
            self.lock()
            return getpass(*args)
        finally:
            self.unlock()


# Global UI object for other modules to use.
ui = UI()



"""Tests for runway.cfngin.hooks.ecs."""
import unittest

import boto3
from moto import mock_ecs
from testfixtures import LogCapture

from runway.cfngin.hooks.ecs import create_clusters

from ..factories import mock_context, mock_provider

REGION = "us-east-1"


class TestECSHooks(unittest.TestCase):
    """Tests for runway.cfngin.hooks.ecs."""

    def setUp(self):
        """Run before tests."""
        self.provider = mock_provider(region=REGION)
        self.context = mock_context(namespace="fake")

    def test_create_single_cluster(self):
        """Test create single cluster."""
        with mock_ecs():
            cluster = "test-cluster"
            logger = "runway.cfngin.hooks.ecs"
            client = boto3.client("ecs", region_name=REGION)
            response = client.list_clusters()

            self.assertEqual(len(response["clusterArns"]), 0)
            with LogCapture(logger) as logs:
                self.assertTrue(
                    create_clusters(
                        provider=self.provider,
                        context=self.context,
                        clusters=cluster,
                    )
                )

                logs.check(
                    (
                        logger,
                        "DEBUG",
                        "Creating ECS cluster: %s" % cluster
                    )
                )

            response = client.list_clusters()
            self.assertEqual(len(response["clusterArns"]), 1)

    def test_create_multiple_clusters(self):
        """Test create multiple clusters."""
        with mock_ecs():
            clusters = ("test-cluster0", "test-cluster1")
            logger = "runway.cfngin.hooks.ecs"
            client = boto3.client("ecs", region_name=REGION)
            response = client.list_clusters()

            self.assertEqual(len(response["clusterArns"]), 0)
            for cluster in clusters:
                with LogCapture(logger) as logs:
                    self.assertTrue(
                        create_clusters(
                            provider=self.provider,
                            context=self.context,
                            clusters=cluster,
                        )
                    )

                    logs.check(
                        (
                            logger,
                            "DEBUG",
                            "Creating ECS cluster: %s" % cluster
                        )
                    )

            response = client.list_clusters()
            self.assertEqual(len(response["clusterArns"]), 2)

    def test_fail_create_cluster(self):
        """Test fail create cluster."""
        with mock_ecs():
            logger = "runway.cfngin.hooks.ecs"
            client = boto3.client("ecs", region_name=REGION)
            response = client.list_clusters()

            self.assertEqual(len(response["clusterArns"]), 0)
            with LogCapture(logger) as logs:
                create_clusters(
                    provider=self.provider,
                    context=self.context
                )

                logs.check(
                    (
                        logger,
                        "ERROR",
                        "setup_clusters hook missing \"clusters\" argument"
                    )
                )

            response = client.list_clusters()
            self.assertEqual(len(response["clusterArns"]), 0)



"""CFNgin entrypoint."""
import logging
import os
import re
import sys

from yaml.constructor import ConstructorError

from runway.util import AWS_ENV_VARS, MutableMap, cached_property

from .actions import build, destroy, diff
from .config import render_parse_load as load_config
from .context import Context as CFNginContext
from .environment import parse_environment
from .providers.aws.default import ProviderBuilder

# explicitly name logger so its not redundant
LOGGER = logging.getLogger('runway.cfngin')


class CFNgin(object):
    """Control CFNgin.

    Attributes:
        EXCLUDE_REGEX (str): Regex used to exclude YAML files when searching
            for config files.
        EXCLUDE_LIST (str): Global list of YAML file names to exclude when
            searching for config files.
        concurrency (int): Max number of CFNgin stacks that can be deployed
            concurrently. If the value is ``0``, will be constrained based on
            the underlying graph.
        interactive (bool): Wether or not to prompt the user before taking
            action.
        parameters (MutableMap): Combination of the parameters provided when
            initalizing the class and any environment files that are found.
        recreate_failed (bool): Destroy and re-create stacks that are stuck in
            a failed state from an initial deployment when updating.
        region (str): The AWS region where CFNgin is currently being executed.
        sys_path (str): Working directory.
        tail (bool): Wether or not to display all CloudFormation events in the
            terminal.

    """

    EXCLUDE_REGEX = r'runway(\..*)?\.(yml|yaml)'
    EXCLUDE_LIST = ['buildspec.yml', 'docker-compose.yml']

    def __init__(self, ctx, parameters=None, sys_path=None):
        """Instantiate class.

        Args:
            ctx (runway.context.Context): Runway context object.
            parameters (Optional[Dict[str. Any]]): Parameters from Runway.
            sys_path (Optional[str]): Working directory.

        """
        self.__ctx = ctx
        self._aws_credential_backup = {}
        self._env_file_name = None
        self.concurrency = ctx.max_concurrent_cfngin_stacks
        self.interactive = ctx.is_interactive
        self.parameters = MutableMap()
        self.recreate_failed = ctx.is_noninteractive
        self.region = ctx.env_region
        self.sys_path = sys_path or os.getcwd()
        self.tail = ctx.debug

        self.parameters.update(self.env_file)

        if parameters:
            LOGGER.debug('Adding Runway parameters to CFNgin parameters')
            self.parameters.update(parameters)

        self._inject_common_parameters()
        self._save_aws_credentials()

    @cached_property
    def env_file(self):
        """Contents of a CFNgin environment file.

        Returns:
            MutableMap

        """
        result = {}
        supported_names = ['{}.env'.format(self.__ctx.env_name),
                           '{}-{}.env'.format(self.__ctx.env_name,
                                              self.region)]
        for _, file_name in enumerate(supported_names):
            file_path = os.path.join(self.sys_path, file_name)
            if os.path.isfile(file_path):
                LOGGER.info('Found environment file: %s', file_path)
                self._env_file_name = file_path
                with open(file_path, 'r') as file_:
                    result.update(parse_environment(file_.read()))
        return MutableMap(**result)

    def deploy(self, force=False, sys_path=None):
        """Run the CFNgin deploy action.

        Args:
            force (bool): Explicitly enable the action even if an environment
                file is not found.
            syspath (Optional[str]): Explicitly define a path to work in.
                If not provided, ``self.sys_path`` is used.

        """
        if self.should_skip(force):
            return
        self._inject_aws_credentials()
        if not sys_path:
            sys_path = self.sys_path
        config_files = self.find_config_files(sys_path=sys_path)

        for config in config_files:
            ctx = self.load(config)
            LOGGER.info('%s: deploying...', os.path.basename(config))
            action = build.Action(
                context=ctx,
                provider_builder=self._get_provider_builder(
                    ctx.config.service_role
                )
            )
            action.execute(concurrency=self.concurrency,
                           tail=self.tail)
        self._restore_aws_credentials()

    def destroy(self, force=False, sys_path=None):
        """Run the CFNgin destroy action.

        Args:
            force (bool): Explicitly enable the action even if an environment
                file is not found.
            syspath (Optional[str]): Explicitly define a path to work in.
                If not provided, ``self.sys_path`` is used.

        """
        if self.should_skip(force):
            return
        self._inject_aws_credentials()
        if not sys_path:
            sys_path = self.sys_path
        config_files = self.find_config_files(sys_path=sys_path)
        # destroy should run in reverse to handle dependencies
        config_files.reverse()

        for config in config_files:
            ctx = self.load(config)
            LOGGER.info('%s: destroying...', os.path.basename(config))
            action = destroy.Action(
                context=ctx,
                provider_builder=self._get_provider_builder(
                    ctx.config.service_role
                )
            )
            action.execute(concurrency=self.concurrency,
                           force=True,
                           tail=self.tail)
        self._restore_aws_credentials()

    def load(self, config_path):
        """Load a CFNgin config into a context object.

        Args:
            config_path (str): Valid path to a CFNgin config file.

        Returns:
            :class:`runway.cfngin.context.Context`

        """
        LOGGER.debug('%s: loading...', os.path.basename(config_path))
        try:
            config = self._get_config(config_path)
            context = self._get_context(config, config_path)
            return context
        except ConstructorError as err:
            if err.problem.startswith('could not determine a constructor '
                                      'for the tag \'!'):
                LOGGER.error('"%s" appears to be a CloudFormation template, '
                             'but is located in the top level of a module '
                             'alongside the CloudFormation config files (i.e. '
                             'the file or files indicating the stack names & '
                             'parameters). Please move the template to a '
                             'subdirectory.', config_path)
                sys.exit(1)
            raise

    def plan(self, force=False, sys_path=None):
        """Run the CFNgin plan action.

        Args:
            force (bool): Explicitly enable the action even if an environment
                file is not found.
            syspath (Optional[str]): Explicitly define a path to work in.
                If not provided, ``self.sys_path`` is used.

        """
        if self.should_skip(force):
            return
        self._inject_aws_credentials()
        if not sys_path:
            sys_path = self.sys_path
        config_files = self.find_config_files(sys_path=sys_path)
        for config in config_files:
            ctx = self.load(config)
            LOGGER.info('%s: generating change sets...',
                        os.path.basename(config))
            action = diff.Action(
                context=ctx,
                provider_builder=self._get_provider_builder(
                    ctx.config.service_role
                )
            )
            action.execute()
        self._restore_aws_credentials()

    def should_skip(self, force=False):
        """Determine if action should be taken or not.

        Args:
            force (bool): If ``True``, will always return ``False`` meaning
                the action should not be skipped.

        Returns:
            bool: Skip action or not.

        """
        if force or self.env_file:
            return False
        LOGGER.info('Skipping module; environment not explicitly enabled '
                    'and no environment file found')
        return True

    def _get_config(self, file_path, validate=True):
        """Initialize a CFNgin config object from a file.

        Args:
            file_path (str): Path to the config file to load.
            validate (bool): Validate the loaded config.

        Returns:
            :class:`runway.cfngin.config.Config`

        """
        with open(file_path, 'r') as file_:
            raw_config = file_.read()
        return load_config(raw_config, self.parameters, validate)

    def _get_context(self, config, config_path):
        """Initialize a CFNgin context object.

        Args:
            config (:class:`runway.cfngin.config.Config): CFNgin config object.
            config_path (str): Path to the config file that was provided.

        Returns:
            :class:`runway.cfngin.context.Context`

        """
        return CFNginContext(
            boto3_credentials=self.__ctx.boto3_credentials,
            config=config,
            config_path=config_path,
            environment=self.parameters,
            force_stacks=[],  # placeholder
            region=self.region,
            stack_names=[]  # placeholder
        )

    def _get_provider_builder(self, service_role=None):
        """Initialize provider builder.

        Args:
            service_role (Optional[str]): CloudFormation service role.

        Returns:
            ProviderBuilder

        """
        if self.interactive:
            LOGGER.info('Using interactive AWS provider mode.')
        else:
            LOGGER.info('Using default AWS provider mode.')
        return ProviderBuilder(
            interactive=self.interactive,
            recreate_failed=self.recreate_failed,
            region=self.region,
            service_role=service_role
        )

    # TODO remove after deprecated `get_session` is removed
    def _inject_aws_credentials(self):
        """Change environment AWS credentials."""
        LOGGER.debug('Injecting AWS credentials into the environment before '
                     'running a CFNgin action')
        os.environ.update(self.__ctx.current_aws_creds)

    def _inject_common_parameters(self):
        """Add common parameters if they don't already exist.

        Adding these commonly used parameters will remove the need to add
        lookup support (mainly for environment variable lookups) in places
        such as ``cfngin_bucket``.

        Injected Parameters
        ~~~~~~~~~~~~~~~~~~~

        **environment (str)**
            Taken from the ``DEPLOY_ENVIRONMENT`` environment variable. This
            will the be current Runway environment being processed.

        **region (str)**
            Taken from the ``AWS_REGION`` environment variable. This will be
            the current region being deployed to.

        """
        if not self.parameters.get('environment'):
            self.parameters['environment'] = self.__ctx.env_name
        if not self.parameters.get('region'):
            self.parameters['region'] = self.region

    # TODO remove after deprecated `get_session` is removed
    def _save_aws_credentials(self):
        """Save a copy of existing AWS credentials from the environment."""
        LOGGER.debug('Creating a backup of AWS credentials from the '
                     'environment before running CFNgin')
        temp = os.environ.copy()
        for name in AWS_ENV_VARS:
            value = temp.get(name)
            if value:
                self._aws_credential_backup[name] = value

    # TODO remove after deprecated `get_session` is removed
    def _restore_aws_credentials(self):
        """Restore original AWS credentials to the environment."""
        LOGGER.debug('Restoring AWS credential backup post CFNgin action')
        for name in AWS_ENV_VARS:
            value = self._aws_credential_backup.get(name)
            if value:
                os.environ[name] = value
            else:
                # remove key if it exists, no error raised if it doesn't
                os.environ.pop(name, None)

    @classmethod
    def find_config_files(cls, exclude=None, sys_path=None):
        """Find CFNgin config files.

        Args:
            exclude (Optional[List[str]]): List of file names to exclude. This
                list is appended to the global exclude list.
            sys_path (Optional[str]): Explicitly define a path to search for
                config files.

        Returns:
            List[str]: Path to config files that were found.

        """
        if not sys_path:
            sys_path = os.getcwd()
        elif os.path.isfile(sys_path):
            return [sys_path]

        exclude = exclude or []
        result = []
        exclude.extend(cls.EXCLUDE_LIST)
        for root, _dirs, files in os.walk(sys_path):
            for name in files:
                if re.match(cls.EXCLUDE_REGEX, name) or (name in exclude or
                                                         name.startswith('.')):
                    # Hidden files (e.g. .gitlab-ci.yml), Runway configs,
                    # and docker-compose files definitely aren't stacker
                    # config files
                    continue
                if os.path.splitext(name)[-1] in ['.yaml', '.yml']:
                    result.append(os.path.join(root, name))
            break  # only need top level files
        result.sort()
        return result




"""Check the authorization information passed along via the cookie headers.

When the information is not present or an error occurs due to verification a new
request to Cognito will be made to authorize the user. The user will be taken
to the Cognito login page to enter their information, and so long as it is valid,
the information will be passed to the parsing agent.

If a refresh token exists and it has expired (after 1 hour) do an automatic refresh
of the credentials by redirecting the user to the refresh agent.
"""
import base64
import datetime
import hashlib
import json
import logging
import secrets  # pylint: disable=import-error
import re
from urllib.parse import quote_plus, urlencode  # noqa pylint: disable=no-name-in-module, import-error

from jose import jwt  # pylint: disable=import-error

from jwks_rsa.client import JwksClient  # pylint: disable=import-error
from shared import decode_token, extract_and_parse_cookies, get_config  # noqa pylint: disable=import-error

LOGGER = logging.getLogger(__file__)

SECRET_ALLOWED_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~"
NONCE_LENGTH = 16
PKCE_LENGTH = 43
CONFIG = get_config()


def handler(event, _context):
    """Handle the request passed in.

    Keyword Args:
        event (Dict[str, Any]): The Lambda Event
    """
    request = event['Records'][0]['cf']['request']
    domain_name = request['headers']['host'][0]['value']
    querystring = request.get('querystring')
    request_query_string = (
        "?%s" % querystring
    ) if querystring else ""
    requested_uri = '%s%s' % (request['uri'], request_query_string)
    nonce = generate_nonce()

    try:
        # Extract the cookies received from Cognito
        extracted = extract_and_parse_cookies(
            request['headers'],
            CONFIG['client_id']
        )
        token_user_name = extracted.get('tokenUserName')
        id_token = extracted.get('idToken')
        refresh_token = extracted.get('refreshToken')

        # If the token user name or id token are missing then we need
        # new credentials
        if not token_user_name or not id_token:
            msg = "No valid credentials present in cookies"
            LOGGER.error(msg)
            raise Exception(msg)

        # Get the expiration date from the id token.
        decoded_token = decode_token(id_token)
        expiration = decoded_token.get('exp')
        exp_date = datetime.datetime.fromtimestamp(expiration)
        now = datetime.datetime.now()

        # If we have a refresh token and the expiration date has passed
        # then forward the user to the refresh agent
        if now > exp_date and refresh_token:
            headers = {
                # Redirect the user to the refresh agent
                'location': [
                    {
                        'key': 'location',
                        'value': 'https://%s%s?%s' % (
                            domain_name,
                            CONFIG.get('redirect_path_auth_refresh'),
                            urlencode({
                                'requestedUri': requested_uri,
                                'nonce': nonce
                            })
                        )
                    }
                ],
                'set-cookie': [
                    # Set the Nonce Cookie
                    {
                        'key': 'set-cookie',
                        'value': 'spa-auth-edge-nonce=%s; %s' % (
                            quote_plus(nonce),
                            CONFIG.get('cookie_settings').get('nonce')
                        )
                    }
                ]
            }
            # Add all CloudFrontHeaders
            headers.update(CONFIG.get('cloud_front_headers'))
            return {
                'status': '307',
                'statusDescription': 'Temporary Redirect',
                'headers': headers
            }

        # Validate the token information against the Cognito JWKS
        validate_jwt(
            id_token,
            CONFIG.get('token_jwks_uri'),
            CONFIG.get('token_issuer'),
            CONFIG.get('client_id'))

        return request
    # We need new authorization. Get the user over to Cognito
    except Exception: # noqa pylint: disable=broad-except
        pkce, pkce_hash = generate_pkce_verifier()
        payload = {
            'redirect_uri': 'https://%s%s' % (domain_name, CONFIG['redirect_path_sign_in']),
            'response_type': 'code',
            'client_id': CONFIG["client_id"],
            'state': base64.urlsafe_b64encode(
                bytes(json.dumps({
                    "nonce": nonce,
                    "requestedUri": requested_uri
                }).encode())
            ),
            'scope': " ".join(CONFIG['oauth_scopes']),
            'code_challenge_method': "S256",
            'code_challenge': pkce_hash
        }
        login_query_string = urlencode(payload, quote_via=quote_plus)
        headers = CONFIG.get('cloud_front_headers')
        headers['location'] = [
            {
                'key': 'location',
                'value': 'https://%s/oauth2/authorize?%s' % (
                    CONFIG['cognito_auth_domain'],
                    login_query_string
                )
            }
        ]
        headers['set-cookie'] = [
            # Set Nonce Cookie
            {
                'key': 'set-cookie',
                'value': 'spa-auth-edge-nonce=%s; %s' % (
                    quote_plus(nonce),
                    CONFIG.get('cookie_settings', {}).get('nonce')
                )
            },
            # Set PKCE Cookie
            {
                'key': 'set-cookie',
                'value': 'spa-auth-edge-pkce=%s; %s' % (
                    quote_plus(pkce),
                    CONFIG.get('cookie_settings', {}).get('nonce')
                )
            },
        ]

        # Redirect user to the Cognito Login
        return {
            'status': '307',
            'statusDescription': 'Temporary Redirect',
            'headers': headers
        }


def generate_pkce_verifier():
    """Generate the PKCE verification code."""
    pkce = random_key(PKCE_LENGTH)
    pkce_hash = hashlib.sha256()
    pkce_hash.update(pkce.encode('UTF-8'))
    pkce_hash = pkce_hash.digest()
    pkce_hash_b64 = base64.b64encode(pkce_hash)
    decoded_pkce = pkce_hash_b64.decode("UTF-8")
    decoded_pkce = re.sub(r'\=', '', decoded_pkce)
    decoded_pkce = re.sub(r'\+', '-', decoded_pkce)
    decoded_pkce = re.sub(r'\/', '_', decoded_pkce)

    return [pkce, decoded_pkce]


def generate_nonce():
    """Generate a random Nonce token."""
    return random_key(NONCE_LENGTH)


def random_key(length=15):
    """Generate a random key of specified length from the allowed secret characters.

    Keyword Args:
        length (int): The length of the random key
    """
    return ''.join(secrets.choice(SECRET_ALLOWED_CHARS) for _ in range(length))


def validate_jwt(jwt_token,
                 jwks_uri,
                 issuer,
                 audience
                ):  # noqa: E124
    """Validate the JWT token against the Cognito JWKs.

    Keyword Args:
        jwt_token (str): The JSON Web Token to validate
        jwks_uri (str): The URI in which to retrieve the JSON Web Keys
        issuer (str): Issuer of the JWT
        audience (str): Audience of the JWT
    """
    token_headers = jwt.get_unverified_header(jwt_token)
    if not token_headers:
        raise Exception('Cannot Parse JWT Token Headers')

    kid = token_headers.get('kid')
    jwk = get_signing_key(jwks_uri, kid)

    return jwt.decode(
        jwt_token,
        jwk,
        algorithms=['RS256'],
        audience=audience,
        issuer=issuer,
        options={'verify_at_hash': False})


def is_rsa_signing_key(key):
    """Verify if the key specified is an RSA Public Key.

    Keyword Args:
        key (Dict): The key to filter
    """
    return "rsaPublicKey" in key


def get_signing_key(jwks_uri, kid):
    """Retrieve the signing keys from the JWKS uri that match the key id specified.

    Keyword Args:
        jwks_uri (str): The URI in which to retrieve the JWKs
        kid (str): Key ID of the signing key we are looking for
    """
    client = JwksClient({'jwks_uri': jwks_uri})
    jwk = client.get_signing_key(kid)
    return jwk.get('rsaPublicKey') if is_rsa_signing_key(jwk) else jwk.get('publicKey')



"""Tests for runway.cfngin.dag."""
import threading

import pytest

from runway.cfngin.dag import (DAGValidationError, ThreadedWalker,
                               UnlimitedSemaphore)


def test_add_node(empty_dag):
    """Test add node."""
    dag = empty_dag

    dag.add_node('a')
    assert dag.graph == {'a': set()}


def test_transpose(basic_dag):
    """Test transpose."""
    dag = basic_dag

    transposed = dag.transpose()
    assert transposed.graph == {'d': set(['c', 'b']),
                                'c': set(['a']),
                                'b': set(['a']),
                                'a': set([])}


def test_add_edge(empty_dag):
    """Test add edge."""
    dag = empty_dag

    dag.add_node('a')
    dag.add_node('b')
    dag.add_edge('a', 'b')
    assert dag.graph == {'a': set('b'), 'b': set()}


def test_from_dict(empty_dag):
    """Test from dict."""
    dag = empty_dag

    dag.from_dict({'a': ['b', 'c'],
                   'b': ['d'],
                   'c': ['d'],
                   'd': []})
    assert dag.graph == {'a': set(['b', 'c']),
                         'b': set('d'),
                         'c': set('d'),
                         'd': set()}


def test_reset_graph(empty_dag):
    """Test reset graph."""
    dag = empty_dag

    dag.add_node('a')
    assert dag.graph == {'a': set()}
    dag.reset_graph()
    assert dag.graph == {}


def test_walk(empty_dag):
    """Test walk."""
    dag = empty_dag

    # b and c should be executed at the same time.
    dag.from_dict({'a': ['b', 'c'],
                   'b': ['d'],
                   'c': ['d'],
                   'd': []})

    nodes = []

    def walk_func(node):
        nodes.append(node)
        return True

    dag.walk(walk_func)
    assert nodes == ['d', 'c', 'b', 'a'] or nodes == ['d', 'b', 'c', 'a']


def test_ind_nodes(basic_dag):
    """Test ind nodes."""
    dag = basic_dag
    assert dag.ind_nodes() == ['a']


def test_topological_sort(empty_dag):
    """Test topological sort."""
    dag = empty_dag
    dag.from_dict({'a': [],
                   'b': ['a'],
                   'c': ['b']})
    assert dag.topological_sort() == ['c', 'b', 'a']


def test_successful_validation(basic_dag):
    """Test successful validation."""
    dag = basic_dag
    assert dag.validate()[0] is True


def test_failed_validation(empty_dag):
    """Test failed validation."""
    dag = empty_dag

    with pytest.raises(DAGValidationError):
        dag.from_dict({'a': ['b'],
                       'b': ['a']})


def test_downstream(basic_dag):
    """Test downstream."""
    dag = basic_dag
    assert set(dag.downstream('a')) == set(['b', 'c'])


def test_all_downstreams(basic_dag):
    """Test all downstreams."""
    dag = basic_dag

    assert dag.all_downstreams('a') == ['b', 'c', 'd']
    assert dag.all_downstreams('b') == ['d']
    assert dag.all_downstreams('d') == []


def test_all_downstreams_pass_graph(empty_dag):
    """Test all downstreams pass graph."""
    dag = empty_dag
    dag.from_dict({'a': ['c'],
                   'b': ['d'],
                   'c': ['d'],
                   'd': []})
    assert dag.all_downstreams('a') == ['c', 'd']
    assert dag.all_downstreams('b') == ['d']
    assert dag.all_downstreams('d') == []


def test_predecessors(basic_dag):
    """Test predecessors."""
    dag = basic_dag

    assert set(dag.predecessors('a')) == set([])
    assert set(dag.predecessors('b')) == set(['a'])
    assert set(dag.predecessors('c')) == set(['a'])
    assert set(dag.predecessors('d')) == set(['b', 'c'])


def test_filter(basic_dag):
    """Test filter."""
    dag = basic_dag

    dag2 = dag.filter(['b', 'c'])
    assert dag2.graph == {'b': set('d'),
                          'c': set('d'),
                          'd': set()}


def test_all_leaves(basic_dag):
    """Test all leaves."""
    dag = basic_dag

    assert dag.all_leaves() == ['d']


def test_size(basic_dag):
    """Test size."""
    dag = basic_dag

    assert dag.size() == 4
    dag.delete_node('a')
    assert dag.size() == 3


def test_transitive_reduction_no_reduction(empty_dag):
    """Test transitive reduction no reduction."""
    dag = empty_dag
    dag.from_dict({'a': ['b', 'c'],
                   'b': ['d'],
                   'c': ['d'],
                   'd': []})
    dag.transitive_reduction()
    assert dag.graph == {'a': set(['b', 'c']),
                         'b': set('d'),
                         'c': set('d'),
                         'd': set()}


def test_transitive_reduction(empty_dag):
    """Test transitive reduction."""
    dag = empty_dag
    # https://en.wikipedia.org/wiki/Transitive_reduction#/media/File:Tred-G.svg
    dag.from_dict({'a': ['b', 'c', 'd', 'e'],
                   'b': ['d'],
                   'c': ['d', 'e'],
                   'd': ['e'],
                   'e': []})
    dag.transitive_reduction()
    # https://en.wikipedia.org/wiki/Transitive_reduction#/media/File:Tred-Gprime.svg
    assert dag.graph == {'a': set(['b', 'c']),
                         'b': set('d'),
                         'c': set('d'),
                         'd': set('e'),
                         'e': set()}


def test_transitive_deep_reduction(empty_dag):
    """Test transitive deep reduction."""
    dag = empty_dag
    # https://en.wikipedia.org/wiki/Transitive_reduction#/media/File:Tred-G.svg
    dag.from_dict({
        'a': ['b', 'd'],
        'b': ['c'],
        'c': ['d'],
        'd': [],
    })
    dag.transitive_reduction()
    # https://en.wikipedia.org/wiki/Transitive_reduction#/media/File:Tred-Gprime.svg
    assert dag.graph == {'a': set('b'),
                         'b': set('c'),
                         'c': set('d'),
                         'd': set()}


def test_threaded_walker(empty_dag):
    """Test threaded walker."""
    dag = empty_dag

    walker = ThreadedWalker(UnlimitedSemaphore())

    # b and c should be executed at the same time.
    dag.from_dict({'a': ['b', 'c'],
                   'b': ['d'],
                   'c': ['d'],
                   'd': []})

    lock = threading.Lock()  # Protects nodes from concurrent access
    nodes = []

    def walk_func(node):
        lock.acquire()
        nodes.append(node)
        lock.release()
        return True

    walker.walk(dag, walk_func)
    assert nodes == ['d', 'c', 'b', 'a'] or nodes == ['d', 'b', 'c', 'a']



"""Hook data lookup."""
# pylint: disable=arguments-differ,unused-argument
import logging
import warnings

from troposphere import BaseAWSObject

from runway.lookups.handlers.base import LookupHandler

LOGGER = logging.getLogger(__name__)
TYPE_NAME = "hook_data"


class HookDataLookup(LookupHandler):
    """Hook data lookup."""

    DEPRECATION_MSG = ('Lookup query syntax "<hook_name>::<key>" has been '
                       'deprecated. Please use the new lookup query syntax.')

    @classmethod
    def legacy_parse(cls, value):
        """Retain support for legacy lookup syntax.

        Args:
            value (str): Parameter(s) given to this lookup.

        Format of value:
            <hook_name>::<key>

        """
        hook_name, key = value.split("::")
        warnings.warn(cls.DEPRECATION_MSG, DeprecationWarning)
        LOGGER.warning(cls.DEPRECATION_MSG)
        return '{}.{}'.format(hook_name, key), {}

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Return the data from ``hook_data``.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        """
        try:
            query, args = cls.parse(value)
        except ValueError:
            query, args = cls.legacy_parse(value)

        result = context.hook_data.find(query, args.get('default'))

        if isinstance(result, BaseAWSObject) and \
                args.get('get') and \
                not args.get('load'):
            args['load'] = 'troposphere'

        if not result:
            raise ValueError('Could not find a value for "%s"' % value)

        if result == args.get('default'):
            # assume default value has already been processed so no need to
            # use these
            args.pop('load', None)
            args.pop('get', None)

        return cls.format_results(result, **args)



#!/usr/bin/env python
"""Module with static website bucket and CloudFront distribution."""
from __future__ import print_function

import hashlib
import logging
import os
from typing import Any, Dict, List, Union  # pylint: disable=unused-import

import awacs.s3
import awacs.sts
import awacs.logs
from awacs.aws import (Action, Allow, Policy, PolicyDocument, Principal,
                       Statement)
from awacs.helpers.trust import make_simple_assume_policy
from troposphere import (AccountId, Join, NoValue, Output, Partition, Region, StackName, Sub, # noqa pylint: disable=unused-import
                         awslambda, cloudfront, iam, s3)

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.context import Context

LOGGER = logging.getLogger('runway')

IAM_ARN_PREFIX = 'arn:aws:iam::aws:policy/service-role/'


class StaticSite(Blueprint):  # pylint: disable=too-few-public-methods
    """Stacker blueprint for creating S3 bucket and CloudFront distribution."""

    VARIABLES = {
        'AcmCertificateArn': {'type': str,
                              'default': '',
                              'description': '(Optional) Cert ARN for site'},
        'Aliases': {'type': list,
                    'default': [],
                    'description': '(Optional) Domain aliases the '
                                   'distribution'},
        'DisableCloudFront': {'type': bool,
                              'default': False,
                              'description': 'Whether to disable CF'},
        'LogBucketName': {'type': str,
                          'default': '',
                          'description': 'S3 bucket for CF logs'},
        'PriceClass': {'type': str,
                       'default': 'PriceClass_100',  # US/Europe
                       'description': 'CF price class for the distribution.'},
        'RewriteDirectoryIndex': {'type': str,
                                  'default': '',
                                  'description': '(Optional) File name to '
                                                 'append to directory '
                                                 'requests.'},
        'RoleBoundaryArn': {'type': str,
                            'default': '',
                            'description': '(Optional) IAM Role permissions '
                                           'boundary applied to any created '
                                           'roles.'},
        'WAFWebACL': {'type': str,
                      'default': '',
                      'description': '(Optional) WAF id to associate with the '
                                     'distribution.'},
        'custom_error_responses': {'type': list,
                                   'default': [],
                                   'description': '(Optional) Custom error '
                                                  'responses.'},
        'lambda_function_associations': {'type': list,
                                         'default': [],
                                         'description': '(Optional) Lambda '
                                                        'function '
                                                        'associations.'},
    }

    @property
    def aliases_specified(self):
        # type: () -> bool
        """Aliases are specified conditional."""
        return self.get_variables()['Aliases'] != ['']

    @property
    def cf_enabled(self):
        # type: () -> bool
        """CloudFront enabled conditional."""
        return not self.get_variables().get('DisableCloudFront', False)

    @property
    def acm_certificate_specified(self):
        # type: () -> bool
        """ACM Certification specified conditional."""
        return self.get_variables()['AcmCertificateArn'] != ''

    @property
    def cf_logging_enabled(self):
        # type: () -> bool
        """CloudFront Logging specified conditional."""
        return self.get_variables()['LogBucketName'] != ''

    @property
    def directory_index_specified(self):
        # type: () -> bool
        """Directory Index specified conditional."""
        return self.get_variables()['RewriteDirectoryIndex'] != ''

    @property
    def role_boundary_specified(self):
        # type: () -> bool
        """IAM Role Boundary specified conditional."""
        return self.get_variables()['RoleBoundaryArn'] != ''

    @property
    def waf_name_specified(self):
        # type: () -> bool
        """WAF name specified conditional."""
        return self.get_variables()['WAFWebACL'] != ''

    def create_template(self):
        # type: () -> None
        """Create template (main function called by Stacker)."""
        self.template.set_version('2010-09-09')
        self.template.set_description('Static Website - Bucket and Distribution')

        # Resources
        bucket = self.add_bucket()

        if self.cf_enabled:
            oai = self.add_origin_access_identity()
            bucket_policy = self.add_cloudfront_bucket_policy(bucket, oai)
            lambda_function_associations = self.get_lambda_associations()

            if self.directory_index_specified:
                index_rewrite = self._get_index_rewrite_role_function_and_version()
                lambda_function_associations = self.get_directory_index_lambda_association(
                    lambda_function_associations,
                    index_rewrite['version']
                )

            distribution_options = self.get_cloudfront_distribution_options(
                bucket,
                oai,
                lambda_function_associations
            )
            distribution = self.add_cloudfront_distribution( # noqa pylint: disable=unused-variable
                bucket_policy,
                distribution_options
            )
        else:
            self.add_bucket_policy(bucket)

    def get_lambda_associations(self):
        # type: () -> List[cloudfront.LambdaFunctionAssociation]
        """Retrieve any lambda associations from the instance variables.

        Keyword Args:
            directory_index_rewrite_version (dict): The directory index rewrite lambda version
                resource

        Return:
            array: Array of lambda function association variables
        """
        variables = self.get_variables()

        # If custom associations defined, use them
        if variables['lambda_function_associations']:
            return [
                cloudfront.LambdaFunctionAssociation(
                    EventType=x['type'],
                    LambdaFunctionARN=x['arn']
                ) for x in variables['lambda_function_associations']
            ]
        return []

    def get_directory_index_lambda_association(  # pylint: disable=no-self-use
            self,
            lambda_associations,  # List[cloudfront.LambdaFunctionAssociation]
            directory_index_rewrite_version):  # awslambda.Version
        # type: (...) ->  List[cloudfront.LambdaFunctionAssociation]
        """Retrieve the directory index lambda associations with the added rewriter.

        Args:
            lambda_associations [List(Any)]: The lambda associations
            directory_index_rewrite_version [Any]: The directory index rewrite version
        """
        lambda_associations.append(
            cloudfront.LambdaFunctionAssociation(
                EventType='origin-request',
                LambdaFunctionARN=directory_index_rewrite_version.ref()
            )
        )
        return lambda_associations

    def get_cloudfront_distribution_options(
            self,
            bucket,  # type: s3.Bucket
            oai,  # type: cloudfront.CloudFrontOriginAccessIdentity
            lambda_function_associations):  # List[cloudfront.LambdaFunctionAssociation]
        # type: (...) -> Dict[str, Any]
        """Retrieve the options for our CloudFront distribution.

        Keyword Args:
            bucket (dict): The bucket resource
            oai (dict): The origin access identity resource
            lambda_function_associations (array): The lambda function association array

        Return:
            dict: The CloudFront Distribution Options

        """
        variables = self.get_variables()
        return {
            'Aliases': self.add_aliases(),
            'Origins': [
                cloudfront.Origin(
                    DomainName=Join(
                        '.',
                        [bucket.ref(),
                         's3.amazonaws.com']),
                    S3OriginConfig=cloudfront.S3OriginConfig(
                        OriginAccessIdentity=Join(
                            '',
                            ['origin-access-identity/cloudfront/',
                             oai.ref()])
                    ),
                    Id='S3Origin'
                )
            ],
            'DefaultCacheBehavior': cloudfront.DefaultCacheBehavior(
                AllowedMethods=['GET', 'HEAD'],
                Compress=False,
                DefaultTTL='86400',
                ForwardedValues=cloudfront.ForwardedValues(
                    Cookies=cloudfront.Cookies(Forward='none'),
                    QueryString=False,
                ),
                LambdaFunctionAssociations=lambda_function_associations,
                TargetOriginId='S3Origin',
                ViewerProtocolPolicy='redirect-to-https'
            ),
            'DefaultRootObject': 'index.html',
            'Logging': self.add_logging_bucket(),
            'PriceClass': variables['PriceClass'],
            'CustomErrorResponses': [
                cloudfront.CustomErrorResponse(
                    ErrorCode=response['ErrorCode'],
                    ResponseCode=response['ResponseCode'],
                    ResponsePagePath=response['ResponsePagePath']
                ) for response in variables['custom_error_responses']
            ],
            'Enabled': True,
            'WebACLId': self.add_web_acl(),
            'ViewerCertificate': self.add_acm_cert()
        }

    def add_aliases(self):
        # type: () -> Union[List[str], NoValue]
        """Add aliases."""
        if self.aliases_specified:
            return self.get_variables()['Aliases']
        return NoValue

    def add_web_acl(self):
        # type: () -> Union[str, NoValue]
        """Add Web ACL."""
        if self.waf_name_specified:
            return self.get_variables()['WAFWebACL']
        return NoValue

    def add_logging_bucket(self):
        # type: () -> Union[cloudfront.Logging, NoValue]
        """Add Logging Bucket."""
        if self.cf_logging_enabled:
            return cloudfront.Logging(
                Bucket=Join('.',
                            [self.get_variables()['LogBucketName'],
                             's3.amazonaws.com'])
            )
        return NoValue

    def add_acm_cert(self):
        # type: () -> Union[cloudfront.ViewerCertificate, NoValue]
        """Add ACM cert."""
        if self.acm_certificate_specified:
            return cloudfront.ViewerCertificate(
                AcmCertificateArn=self.get_variables()['AcmCertificateArn'],
                SslSupportMethod='sni-only'
            )
        return NoValue

    def add_origin_access_identity(self):
        # type: () -> cloudfront.CloudFrontOriginAccessIdentity
        """Add the origin access identity resource to the template.

        Returns:
            dict: The OAI resource

        """
        return self.template.add_resource(
            cloudfront.CloudFrontOriginAccessIdentity(
                'OAI',
                CloudFrontOriginAccessIdentityConfig=cloudfront.CloudFrontOriginAccessIdentityConfig(  # noqa pylint: disable=line-too-long
                    Comment='CF access to website'
                )
            )
        )

    def add_bucket_policy(self, bucket):
        # type: (s3.Bucket) -> s3.BucketPolicy
        """Add a policy to the bucket if CloudFront is disabled. Ensure PublicRead.

        Keyword Args:
            bucket (dict): The bucket resource to place the policy

        Returns:
            dict: The Bucket Policy Resource

        """
        return self.template.add_resource(
            s3.BucketPolicy(
                'BucketPolicy',
                Bucket=bucket.ref(),
                PolicyDocument=Policy(
                    Version="2012-10-17",
                    Statement=[
                        Statement(
                            Effect=Allow,
                            Principal=Principal('*'),
                            Action=[Action('s3', 'getObject')],
                            Resource=[
                                Join('', [bucket.get_att('Arn'), '/*'])
                            ],
                        )
                    ]
                )
            )
        )

    def add_bucket(self):
        # type: () -> s3.Bucket
        """Add the bucket resource along with an output of it's name / website url.

        Returns:
            dict: The bucket resource

        """
        bucket = self.template.add_resource(
            s3.Bucket(
                'Bucket',
                AccessControl=(s3.Private if self.cf_enabled else s3.PublicRead),
                LifecycleConfiguration=s3.LifecycleConfiguration(
                    Rules=[
                        s3.LifecycleRule(
                            NoncurrentVersionExpirationInDays=90,
                            Status='Enabled'
                        )
                    ]
                ),
                VersioningConfiguration=s3.VersioningConfiguration(
                    Status='Enabled'
                ),
                WebsiteConfiguration=s3.WebsiteConfiguration(
                    IndexDocument='index.html',
                    ErrorDocument='error.html'
                )
            )
        )
        self.template.add_output(Output(
            'BucketName',
            Description='Name of website bucket',
            Value=bucket.ref()
        ))

        if not self.cf_enabled:
            self.template.add_output(Output(
                'BucketWebsiteURL',
                Description='URL of the bucket website',
                Value=bucket.get_att('WebsiteURL')
            ))

        return bucket

    def add_cloudfront_bucket_policy(self, bucket, oai):
        # type (s3.Bucket, cloudfront.CloudFrontOriginAccessIdentity) -> s3.BucketPolicy
        """Given a bucket and oai resource add cloudfront access to the bucket.

        Keyword Args:
            bucket (dict): A bucket resource
            oai (dict): An Origin Access Identity resource

        Return:
            dict: The CloudFront Bucket access resource
        """
        return self.template.add_resource(
            s3.BucketPolicy(
                'AllowCFAccess',
                Bucket=bucket.ref(),
                PolicyDocument=PolicyDocument(
                    Version='2012-10-17',
                    Statement=self._get_cloudfront_bucket_policy_statements(
                        bucket,
                        oai
                    )
                )
            )
        )

    def add_lambda_execution_role(self,
                                  name='LambdaExecutionRole',  # type: str
                                  function_name=''
                                 ):  # noqa: E124
        # type: (...) -> iam.Role
        """Create the Lambda@Edge execution role."""
        variables = self.get_variables()

        lambda_resource = Join('', [
            'arn:',
            Partition,
            ':logs:*:',
            AccountId,
            ':log-group:/aws/lambda/',
            StackName,
            '-%s-*' % function_name
        ])

        edge_resource = Join('', [
            'arn:',
            Partition,
            ':logs:*:',
            AccountId,
            ':log-group:/aws/lambda/*.',
            StackName,
            '-%s-*' % function_name,
        ])

        return self.template.add_resource(
            iam.Role(
                name,
                AssumeRolePolicyDocument=make_simple_assume_policy(
                    'lambda.amazonaws.com', 'edgelambda.amazonaws.com'
                ),
                PermissionsBoundary=(
                    variables['RoleBoundaryArn'] if self.role_boundary_specified
                    else NoValue
                ),
                Policies=[
                    iam.Policy(
                        PolicyName="LambdaLogCreation",
                        PolicyDocument=PolicyDocument(
                            Version='2012-10-17',
                            Statement=[
                                Statement(
                                    Action=[awacs.logs.CreateLogGroup,
                                            awacs.logs.CreateLogStream,
                                            awacs.logs.PutLogEvents],
                                    Effect=Allow,
                                    Resource=[lambda_resource, edge_resource]
                                )
                            ]
                        )
                    ),
                ],
            )
        )

    def add_cloudfront_directory_index_rewrite(self, role):
        # type: (iam.Role) -> awslambda.Function
        """Add an index CloudFront directory index rewrite lambda function to the template.

        Keyword Args:
            role (dict): The index rewrite role resource

        Return:
            dict: The CloudFront directory index rewrite lambda function resource
        """
        variables = self.get_variables()
        code_str = ''
        path = os.path.join(
            os.path.dirname(__file__),
            'templates/cf_directory_index_rewrite.template.js'
        )
        with open(path) as file_:
            code_str = file_.read().replace(
                '{{RewriteDirectoryIndex}}',
                variables["RewriteDirectoryIndex"]
            )

        return self.template.add_resource(
            awslambda.Function(
                'CFDirectoryIndexRewrite',
                Code=awslambda.Code(
                    ZipFile=code_str
                ),
                Description='Rewrites CF directory HTTP requests to default page',  # noqa
                Handler='index.handler',
                Role=role.get_att('Arn'),
                Runtime='nodejs10.x'
            )
        )

    def add_cloudfront_directory_index_rewrite_version(self, directory_index_rewrite):
        # type: (awslambda.Function) -> awslambda.Version
        """Add a specific version to the directory index rewrite lambda.

        Keyword Args:
            directory_index_rewrite (dict): The directory index rewrite lambda resource

        Return:
            dict: The CloudFront directory index rewrite version

        """
        code_hash = hashlib.md5(
            str(directory_index_rewrite.properties['Code'].properties['ZipFile']).encode()  # noqa pylint: disable=line-too-long
        ).hexdigest()

        return self.template.add_resource(
            awslambda.Version(
                'CFDirectoryIndexRewriteVer' + code_hash,
                FunctionName=directory_index_rewrite.ref()
            )
        )

    def add_cloudfront_distribution(
            self,
            bucket_policy,
            cloudfront_distribution_options
    ):
        # type: (s3.BucketPolicy, Dict[str, Any]) -> cloudfront.Distribution
        """Add the CloudFront distribution to the template / output the id and domain name.

        Keyword Args:
            bucket_policy (dict): Bucket policy to allow CloudFront access
            cloudfront_distribution_options (dict): The distribution options

        Return:
            dict: The CloudFront Distribution resource

        """
        distribution = self.template.add_resource(
            cloudfront.Distribution(
                'CFDistribution',
                DependsOn=bucket_policy.title,
                DistributionConfig=cloudfront.DistributionConfig(
                    **cloudfront_distribution_options
                )
            )
        )
        self.template.add_output(Output(
            'CFDistributionId',
            Description='CloudFront distribution ID',
            Value=distribution.ref()
        ))
        self.template.add_output(
            Output(
                'CFDistributionDomainName',
                Description='CloudFront distribution domain name',
                Value=distribution.get_att('DomainName')
            )
        )
        return distribution

    def _get_cloudfront_bucket_policy_statements(self, bucket, oai):  # noqa pylint: disable=no-self-use
        return [
            Statement(
                Action=[awacs.s3.GetObject],
                Effect=Allow,
                Principal=Principal(
                    'CanonicalUser',
                    oai.get_att('S3CanonicalUserId')
                ),
                Resource=[
                    Join('', [bucket.get_att('Arn'), '/*'])
                ]
            )
        ]

    def _get_index_rewrite_role_function_and_version(self):
        res = {}
        res['role'] = self.add_lambda_execution_role(
            'CFDirectoryIndexRewriteRole',
            'CFDirectoryIndexRewrite'
        )
        res['function'] = self.add_cloudfront_directory_index_rewrite(res['role'])
        res['version'] = self.add_cloudfront_directory_index_rewrite_version(
            res['function']
        )
        return res


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    print(StaticSite('test', Context({"namespace": "test"}), None).to_json())



"""Tests for runway.cfngin.lookups.handlers.default."""
import unittest

from mock import MagicMock

from runway.cfngin.context import Context
from runway.cfngin.lookups.handlers.default import DefaultLookup


class TestDefaultLookup(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.default.DefaultLookup."""

    def setUp(self):
        """Run before tests."""
        self.provider = MagicMock()
        self.context = Context(
            environment={
                'namespace': 'test',
                'env_var': 'val_in_env'}
        )

    def test_env_var_present(self):
        """Test env var present."""
        lookup_val = "env_var::fallback"
        value = DefaultLookup.handle(lookup_val,
                                     provider=self.provider,
                                     context=self.context)
        assert value == 'val_in_env'

    def test_env_var_missing(self):
        """Test env var missing."""
        lookup_val = "bad_env_var::fallback"
        value = DefaultLookup.handle(lookup_val,
                                     provider=self.provider,
                                     context=self.context)
        assert value == 'fallback'

    def test_invalid_value(self):
        """Test invalid value."""
        value = "env_var:fallback"
        with self.assertRaises(ValueError):
            DefaultLookup.handle(value,
                                 provider=self.provider,
                                 context=self.context)



"""AWS SSM Parameter Store lookup."""
# pylint: disable=arguments-differ,unused-argument
import logging
import warnings

from runway.lookups.handlers.base import LookupHandler

from ...session_cache import get_session
from ...util import read_value_from_path

LOGGER = logging.getLogger(__name__)
TYPE_NAME = "ssmstore"


class SsmstoreLookup(LookupHandler):
    """AWS SSM Parameter Store lookup."""

    DEPRECATION_MSG = ('The "ssmstore" lookup has been deprecated. '
                       'The "ssm" lookup should be used instead.')

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Retrieve (and decrypt) a parameter from AWS SSM Parameter Store.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        Returns:
            str: Looked up value.

        ``value`` should be in the following format::

            [<region>@]ssmkey

        .. note:: The region is optional, and defaults to us-east-1 if not given.

        Example:
            ::

                # In CFNgin we would reference the encrypted value like:
                conf_key: ${ssmstore us-east-1@ssmkey}

            You can optionally store the value in a file, ie::

                ssmstore_value.txt
                us-east-1@ssmkey

            and reference it within CFNgin (NOTE: the path should be relative
            to the CFNgin config file)::

                conf_key: ${ssmstore file://ssmstore_value.txt}

                # Both of the above would resolve to
                conf_key: PASSWORD

        """
        warnings.warn(cls.DEPRECATION_MSG, DeprecationWarning)
        LOGGER.warning(cls.DEPRECATION_MSG)

        value = read_value_from_path(value)

        region = "us-east-1"
        if "@" in value:
            region, value = value.split("@", 1)

        client = get_session(region).client("ssm")
        response = client.get_parameters(
            Names=[
                value,
            ],
            WithDecryption=True
        )
        if 'Parameters' in response:
            return str(response['Parameters'][0]['Value'])

        raise ValueError('SSMKey "{}" does not exist in region {}'.format(
            value, region))



"""User Pool Client Domain Updater."""
import logging
from typing import Any, Dict, Optional, Union  # pylint: disable=unused-import

from runway.cfngin.context import Context  # pylint: disable=unused-import
from runway.cfngin.providers.base import BaseProvider  # pylint: disable=unused-import
from runway.cfngin.session_cache import get_session

LOGGER = logging.getLogger(__name__)


def update(context,  # type: Context # pylint: disable=unused-import
           provider,  # type: BaseProvider
           **kwargs  # type: Optional[Dict[str, Any]]
          ):  # noqa: E124
    # type: (...) -> Union[Dict[str, Any], bool]
    """Retrieve/Update the domain name of the specified client.

    A domain name is required in order to make authorization and token
    requests. This prehook ensures we have one available, and if not
    we create one based on the user pool and client ids.

    Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance

    Keyword Args:
        user_pool_id (str): The ID of the Cognito User Pool
        client_id (str): The ID of the Cognito User Pool Client
    """
    session = get_session(provider.region)
    cognito_client = session.client('cognito-idp')

    context_dict = {}

    user_pool_id = context.hook_data['aae_user_pool_id_retriever']['id']
    client_id = kwargs['client_id']
    user_pool = cognito_client.describe_user_pool(UserPoolId=user_pool_id).get('UserPool')
    (user_pool_region, user_pool_hash) = user_pool_id.split('_')

    domain_prefix = user_pool.get('CustomDomain') or user_pool.get('Domain')

    # Return early if we already have a domain
    if domain_prefix:
        context_dict['domain'] = get_user_pool_domain(
            domain_prefix,
            user_pool_region
        )
        return context_dict

    try:
        domain_prefix = ('%s-%s' % (user_pool_hash, client_id)).lower()

        cognito_client.create_user_pool_domain(
            Domain=domain_prefix,
            UserPoolId=user_pool_id
        )
        context_dict['domain'] = get_user_pool_domain(
            domain_prefix,
            user_pool_region
        )
        return context_dict
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.error(
            'Could not update user pool domain for user pool id %s.',
            user_pool_id
        )
        LOGGER.error(err)
        return False


def delete(context,  # type: Context # pylint: disable=unused-import
           provider,  # type: BaseProvider
           **kwargs  # type: Optional[Dict[str, Any]]
          ):  # noqa: E124
    # type: (...) -> Union[Dict[str, Any], bool]
    """Delete the domain if the user pool was created by Runway.

    If a User Pool was created by Runway, and populated with a domain, that
    domain must be deleted prior to the User Pool itself being deleted or an
    error will occur. This process ensures that our generated domain name is
    deleted, or skips if not able to find one.

    Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance

    Keyword Args:
        client_id (str): The ID of the Cognito User Pool Client
    """
    session = get_session(provider.region)
    cognito_client = session.client('cognito-idp')

    user_pool_id = context.hook_data['aae_user_pool_id_retriever']['id']
    client_id = kwargs['client_id']
    (_, user_pool_hash) = user_pool_id.split('_')
    domain_prefix = ('%s-%s' % (user_pool_hash, client_id)).lower()

    try:
        cognito_client.delete_user_pool_domain(
            UserPoolId=user_pool_id,
            Domain=domain_prefix
        )
        return True
    except cognito_client.exceptions.InvalidParameterException:
        LOGGER.info('No domain found with prefix %s. Skipping deletion.', domain_prefix)
        return True
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.error('Could not delete the User Pool Domain.')
        LOGGER.error(err)
        return False


def get_user_pool_domain(prefix, region):
    """Return a user pool domain name based on the prefix received and region.

    Args:
        prefix (str): The domain prefix for the domain
        region (str): The region in which the pool resides
    """
    return '%s.auth.%s.amazoncognito.com' % (prefix, region)



"""Shared functionality for the Auth@Edge Lambda suite."""

import base64
import json
import logging
import re
import time
import traceback
from datetime import datetime
from random import random
from urllib import request  # pylint: disable=no-name-in-module
from urllib.parse import urlencode  # pylint: disable=no-name-in-module,import-error

LOGGER = logging.getLogger(__name__)


def get_config():
    """Retrieve the configuration variables for the Auth@Edge suite.

    Lambda@Edge restricts the ability to use environment variables. This configuration
    object is generated with hard coded values via Runway.
    """
    # This configuration will be replaced with dynamic values
    # via the lambda_config.py Runway hook. Please review
    # that file for more information.
    config = {
        'client_id': 'tbd',
        'cognito_auth_domain': 'tbd',
        'cookie_settings': {},
        'http_headers': {},
        'oauth_scopes': [],
        'redirect_path_auth_refresh': 'tbd',
        'redirect_path_sign_in': 'tbd',
        'redirect_path_sign_out': 'tbd',
        'user_pool_id': 'tbd',
    }

    user_pool_region = 'us-east-1'
    region_match = re.match(
        r"^(\\S+?)_\\S+$",
        config['user_pool_id']
    )
    if region_match:
        user_pool_region = region_match.groups()[0]

    config['cloud_front_headers'] = as_cloud_front_headers(
        config['http_headers']
    )
    config['token_issuer'] = 'https://cognito-idp.%s.amazonaws.com/%s' % (
        user_pool_region,
        config['user_pool_id']
    )
    config['token_jwks_uri'] = '%s/.well-known/jwks.json' % config['token_issuer']
    return config


def as_cloud_front_headers(headers):
    """Convert a series of headers to CloudFront compliant ones.

    Keyword Args:
         headers (Dict[str, str]): The request/response headers in
            dictionary format.
    """
    res = {}
    for key, value in headers.items():
        res[key.lower()] = [{'key': key, 'value': value}]
    return res


def extract_and_parse_cookies(headers, client_id):
    """Extract and parse the Cognito cookies from the headers.

    Keyword Args:
         headers (Dict[str, str]): The request/response headers in
            dictionary format.
        client_id (str): The Cognito UserPool Client ID
    """
    cookies = extract_cookies_from_headers(headers)

    if not cookies:
        return {}

    key_prefix = 'CognitoIdentityServiceProvider.%s' % client_id
    last_user_key = '%s.LastAuthUser' % key_prefix
    token_user_name = cookies.get(last_user_key, '')

    scope_key = '%s.%s.tokenScopesString' % (key_prefix, token_user_name)
    scopes = cookies.get(scope_key, '')

    id_token_key = '%s.%s.idToken' % (key_prefix, token_user_name)
    id_token = cookies.get(id_token_key, '')

    access_token_key = '%s.%s.accessToken' % (key_prefix, token_user_name)
    access_token = cookies.get(access_token_key, '')

    refresh_token_key = '%s.%s.refreshToken' % (key_prefix, token_user_name)
    refresh_token = cookies.get(refresh_token_key, '')

    return {
        'tokenUserName': token_user_name,
        'idToken': id_token,
        'accessToken': access_token,
        'refreshToken': refresh_token,
        'scopes': scopes,
        'nonce': cookies.get('spa-auth-edge-nonce', ''),
        'pkce': cookies.get('spa-auth-edge-pkce', '')
    }


def extract_cookies_from_headers(headers):
    """Extract all cookies from the response headers.

    Keyword Args:
         headers (Dict[str, Dict[str, str]]): The request/response headers in
            dictionary format.
    """
    if "cookie" not in headers:
        return {}

    cookies = {}
    for header in headers["cookie"]:
        split = header.get('value', '').split(';')
        for part in split:
            seq = part.split("=")
            key, value = seq[0], seq[1:]
            cookies[key.strip()] = '='.join(value).strip()

    return cookies


def decode_token(jwt):
    """Decode the JWT and load it's respective parts as JSON.

    Keyword Args:
        jwt (str): The JSON Web Token to parse/decode
    """
    token_body = jwt.split('.')[1]
    decodable_token_body = re.sub(r'\-', '+', token_body)
    decodable_token_body = re.sub(r'\_', '/', token_body)
    decodable_token_body = base64.b64decode(decodable_token_body + "===")
    return json.loads(decodable_token_body.decode("utf8").replace("'", '"'))


def with_cookie_domain(distribution_domain_name, cookie_settings):
    """Check to see if the cookie has a domain, if not then add an Amplify JS compatible one.

    Keyword Args:
        distribution_domain_name (str): The domain name of the
            CloudFront distribution
        cookie_settings (str): The other settings for the cookie
    """
    try:
        cookie_settings.lower().index('domain')
        return cookie_settings
    except ValueError:
        # Add leading dot for compatibility with Amplify (js-cookie)
        return '%s; Domain=.%s' % (cookie_settings, distribution_domain_name)


def get_cookie_headers(client_id,
                       oauth_scopes,
                       tokens,
                       domain_name,
                       cookie_settings,
                       expire_all_tokens=False):
    """Retrieve all cookie headers for our request.

    Return as CloudFront formatted headers.

    Keyword Args:
        client_id (str): The Cognito UserPool Client ID
        oauth_scopes (List): The scopes for oauth validation
        tokens (Dict[str, str]): The tokens received from
            the Cognito Request (id, access, refresh)
        domain_name (str): The Domain name the cookies are
            to be associated with
        cookie_settings (Dict[str, str]): The various settings
            that we would like for the various tokens
        expire_all_tokens (Optional[bool]): Whether to expire
            all Cognito tokens
    """
    decoded_id_token = decode_token(tokens['id_token'])
    token_user_name = decoded_id_token.get('cognito:username')
    key_prefix = 'CognitoIdentityServiceProvider.%s' % client_id
    key_and_user_prefix = '%s.%s' % (key_prefix, token_user_name)
    id_token_key = '%s.idToken' % key_and_user_prefix
    access_token_key = '%s.accessToken' % key_and_user_prefix
    refresh_token_key = '%s.refreshToken' % key_and_user_prefix
    last_user_key = '%s.LastAuthUser' % key_prefix
    scope_key = '%s.tokenScopesString' % key_and_user_prefix
    scopes_string = ' '.join(oauth_scopes)
    user_data_key = '%s.userData' % key_and_user_prefix
    user_data = {
        'UserAttributes': [
            {
                'Name': "sub",
                'Value': decoded_id_token.get('sub')
            },
            {
                'Name': 'email',
                'Value': decoded_id_token.get('email')
            }
        ],
        'Username': token_user_name
    }

    cookies = {
        id_token_key: '%s; %s' % (
            tokens.get('id_token'),
            with_cookie_domain(
                domain_name,
                cookie_settings.get('idToken')
            )
        ),
        access_token_key: '%s; %s' % (
            tokens.get('access_token'),
            with_cookie_domain(
                domain_name,
                cookie_settings.get('accessToken')
            )
        ),
        refresh_token_key: '%s; %s' % (
            tokens.get('refresh_token'),
            with_cookie_domain(
                domain_name,
                cookie_settings.get('refreshToken')
            )
        ),
        last_user_key: '%s; %s' % (
            token_user_name,
            with_cookie_domain(
                domain_name,
                cookie_settings.get('idToken')
            )
        ),
        scope_key: '%s; %s' % (
            scopes_string,
            with_cookie_domain(
                domain_name,
                cookie_settings.get('accessToken')
            )
        ),
        user_data_key: '%s; %s' % (
            urlencode(user_data),
            with_cookie_domain(
                domain_name,
                cookie_settings.get('idToken')
            )
        ),
        'amplify-signin-with-hostedUI': 'true; %s' % (
            with_cookie_domain(
                domain_name,
                cookie_settings.get('accessToken')
            )
        )
    }

    if expire_all_tokens:
        for key in cookies:
            cookies[key] = expire_cookie(cookies[key])
    elif not tokens.get('refresh_token'):
        cookies[refresh_token_key] = expire_cookie(cookies[refresh_token_key])

    cloud_front_headers = []
    for key, val in cookies.items():
        cloud_front_headers.append({'key': 'set-cookie', 'value': '%s=%s' % (key, val)})

    return cloud_front_headers


def expire_cookie_filter(cookie):
    """Filter to determine if a cookie starts with 'max-age' or 'expires'.

    Keyword Args:
        cookie (str): The cookie to filter
    """
    if cookie.lower().startswith("max-age") or cookie.lower().startswith('expires'):
        return False
    return True


def expire_cookie(cookie):
    """Add an expiration property to a specific cookie.

    Keyword Arguments:
        cookie (str): The cookie string
    """
    cookie_parts = cookie.split(';')
    mapped = [c.strip() for c in cookie_parts]
    filtered = filter(expire_cookie_filter, mapped)
    listed = list(filtered)
    _, settings = listed[0], listed[1:]
    settings.insert(0, '')
    settings.append("Expires=%s" % datetime.utcnow())
    return "; ".join(settings)


def http_post_with_retry(url, data, headers):
    """Given a URL/Data/Headers make a POST request with exponential backoff.

    Used for retrieving token information from Cognito

    Keyword Args:
        url (str): The URL to make the POST request to
        data (Dict[str, str]): The dictionary of data elements to
            send with the request (urlencoded internally)
        headers (List[Dict[str, str]]): Any headers to send with
            the POST request
    """
    attempts = 1
    while attempts:
        try:
            encoded_data = urlencode(data).encode()
            req = request.Request(url, encoded_data, headers)
            res = request.urlopen(req).read()
            read = res.decode('utf-8')
            json_data = json.loads(read)
            return json_data
        # pylint: disable=broad-except
        except Exception as err:
            LOGGER.error('HTTP POST to %s failed (attempt %s)', url, attempts)
            LOGGER.error(err)
            LOGGER.error(traceback.print_exc())
            if attempts >= 5:
                break
            if attempts >= 2:
                # After attempting twice do some exponential backoff with jitter
                time.sleep(
                    (25 * (pow(2, attempts) + random() * attempts)) / 1000
                )
        finally:
            attempts += 1


def create_error_html(title, message, try_again_href):
    """Create a basic error html page for exception returns.

    Keyword Args:
        title (str): The title of the page
        message (str): Any exception message
        try_again_href (str): URL href to try the request again
    """
    return "<!DOCTYPE html>" + \
           "<html lang='en'>" + \
           "    <head>" + \
           "        <meta charset='utf-8'>" + \
           "        <title>%s</title>" % title + \
           "    </head>" + \
           "    <body>" + \
           "        <h1>%s</h1>" % title + \
           "        <p><b>ERROR:</b> %s</p>" % message + \
           "        <a href='%s'>Try again</a>" % try_again_href + \
           "    </body>" + \
           "</html>"



"""Runway context module."""
import logging
# needed for python2 cpu_count, can be replace with python3 os.cpu_count()
import multiprocessing
import os
import sys

from .util import AWS_ENV_VARS
from .cfngin.session_cache import get_session

LOGGER = logging.getLogger('runway')


def echo_detected_environment(env_name, env_vars):
    """Print a helper note about how the environment was determined."""
    env_override_name = 'DEPLOY_ENVIRONMENT'
    LOGGER.info("")
    if env_override_name in env_vars:
        LOGGER.info("Environment \"%s\" was determined from the %s environment variable.",
                    env_name,
                    env_override_name)
        LOGGER.info("If this is not correct, update "
                    "the value (or unset it to fall back to the name of "
                    "the current git branch or parent directory).")
    else:
        LOGGER.info("Environment \"%s\" was determined from the current "
                    "git branch or parent directory.",
                    env_name)
        LOGGER.info("If this is not the environment name, update the branch/folder name or "
                    "set an override value via the %s environment variable",
                    env_override_name)
    LOGGER.info("")


class Context(object):
    """Runway execution context."""

    env_override_name = 'DEPLOY_ENVIRONMENT'

    def __init__(self, env_name,  # pylint: disable=too-many-arguments
                 env_region, env_root, env_vars=None,
                 command=None):
        """Initialize base class."""
        self.env_name = env_name
        self.env_region = env_region
        self.env_root = env_root
        self.command = command
        self.env_vars = env_vars or os.environ.copy()
        self._env_name_from_env = bool(self.env_vars.get(self.env_override_name))
        self.debug = bool(self.env_vars.get('DEBUG'))

        self.echo_detected_environment()

        if not self._env_name_from_env:
            self.env_vars.update({'DEPLOY_ENVIRONMENT': self.env_name})

    @property
    def boto3_credentials(self):
        """Return a dict of boto3 credentials."""
        return {key.lower(): value
                for key, value in self.current_aws_creds.items()}

    @property
    def current_aws_creds(self):
        """AWS credentials from self.env_vars.

        Returns:
            Dict[str, str]

        """
        return {name: self.env_vars.get(name)
                for name in AWS_ENV_VARS if self.env_vars.get(name)}

    @property
    def is_interactive(self):
        """Wether the user should be prompted or not.

        Determined by the existed of ``CI`` in the environment.

        Returns:
            bool

        """
        return not bool(self.env_vars.get('CI'))

    @property
    def is_noninteractive(self):
        """Wether the user should be prompted or not.

        Determined by the existed of ``CI`` in the environment.
        Inverse of ``is_interactive`` property.

        Returns:
            bool

        """
        return bool(self.env_vars.get('CI'))

    @property
    def is_python3(self):
        """Wether running in Python 3 or not.

        Used for Python compatability decisions.

        Returns:
            bool

        """
        return sys.version_info.major > 2

    @property
    def max_concurrent_cfngin_stacks(self):
        """Max number of CFNgin stacks that can be deployed concurrently.

        This property can be set by exporting
        ``RUNWAY_MAX_CONCURRENT_CFNGIN_STACKS``. If no value is specified, the
        value will be constrained based on the underlying graph.

        Returns:
            int: Value from environment variable or ``0``.

        """
        return int(
            self.env_vars.get('RUNWAY_MAX_CONCURRENT_CFNGIN_STACKS', '0')
        )

    @property
    def max_concurrent_modules(self):
        """Max number of modules that can be deployed to concurrently.

        This property can be set by exporting ``RUNWAY_MAX_CONCURRENT_MODULES``.
        If no value is specified, ``min(61, os.cpu_count())`` is used.

        On Windows, this must be equal to or lower than ``61``.

        **IMPORTANT:** When using ``parallel_regions`` and ``child_modules``
        together, please consider the nature of their relationship when
        manually setting this value. (``parallel_regions * child_modules``)

        Returns:
            int: Value from environment variable or ``min(61, os.cpu_count())``

        """
        value = self.env_vars.get('RUNWAY_MAX_CONCURRENT_MODULES')

        if value:
            return int(value)
        # TODO update to `os.cpu_count()` when dropping python2
        return min(61, multiprocessing.cpu_count())

    @property
    def max_concurrent_regions(self):
        """Max number of regions that can be deployed to concurrently.

        This property can be set by exporting ``RUNWAY_MAX_CONCURRENT_REGIONS``.
        If no value is specified, ``min(61, os.cpu_count())`` is used.

        On Windows, this must be equal to or lower than ``61``.

        **IMPORTANT:** When using ``parallel_regions`` and ``child_modules``
        together, please consider the nature of their relationship when
        manually setting this value. (``parallel_regions * child_modules``)

        Returns:
            int: Value from environment variable or ``min(61, os.cpu_count())``

        """
        value = self.env_vars.get('RUNWAY_MAX_CONCURRENT_REGIONS')

        if value:
            return int(value)
        # TODO update to `os.cpu_count()` when dropping python2
        return min(61, multiprocessing.cpu_count())

    @property
    def use_concurrent(self):
        """Wether to use concurrent.futures or not.

        Noninteractive is required for concurrent execution to prevent weird
        user-input behavior.

        Python 3 is required because backported futures has issues with
        ProcessPoolExecutor.

        Returns:
            bool

        """
        if self.is_noninteractive and self.is_python3:
            return True
        return False

    def echo_detected_environment(self):
        """Print a helper note about how the environment was determined."""
        LOGGER.info("")
        if self._env_name_from_env:
            LOGGER.info("Environment \"%s\" was determined from the %s "
                        "environment variable.", self.env_name,
                        self.env_override_name)
            LOGGER.info("If this is not correct, update "
                        "the value (or unset it to fall back to the name of "
                        "the current git branch or parent directory).")
        else:
            LOGGER.info("Environment \"%s\" was determined from the current "
                        "git branch or parent directory.",
                        self.env_name)
            LOGGER.info("If this is not the environment name, update the "
                        "branch/folder name or set an override value via "
                        "the %s environment variable", self.env_override_name)
        LOGGER.info("")

    def get_session(self, profile=None, region=None):
        """Create a thread-safe boto3 session.

        Args:
            profile (Optional[str]): The profile for the session.
            region (Optional[str]): The region for the session.

        Returns:
            :class:`boto3.session.Session`: A thread-safe boto3 session.

        """
        kwargs = {}
        # save to var so its not calculated multiple times
        creds = self.boto3_credentials
        if profile:
            kwargs['profile'] = profile
        elif creds:
            kwargs.update({
                'access_key': creds.get('aws_access_key_id'),
                'secret_key': creds.get('aws_secret_access_key'),
                'session_token': creds.get('aws_session_token')
            })
            LOGGER.warning('Current env_region: %s', self.env_region)
        return get_session(region=region or self.env_region, **kwargs)

    def save_existing_iam_env_vars(self):
        """Backup IAM environment variables for later restoration."""
        for i in AWS_ENV_VARS:
            if i in self.env_vars:
                self.env_vars['OLD_' + i] = self.env_vars[i]

    def restore_existing_iam_env_vars(self):
        """Restore backed up IAM environment variables."""
        for i in AWS_ENV_VARS:
            if 'OLD_' + i in self.env_vars:
                self.env_vars[i] = self.env_vars['OLD_' + i]
            elif i in self.env_vars:
                self.env_vars.pop(i)



"""Runway configuration 'path' settings."""
# pylint: disable=unused-import
from typing import Dict, List, Optional, Tuple, Union  # noqa: F401

import os
import logging
import six

from .sources.git import Git

LOGGER = logging.getLogger('runway')


class Path(object):  # pylint: disable=too-many-instance-attributes
    """Runway configuration ``path`` settings object.

    Path is responsible for parsing the ``path`` property of a Runway
    configuration. It then can determine if the path specified is
    locally sourced or remotely sourced through a service such
    as `Git`_ or S3.

    Local ``path`` variables are defined relative to the root project folder.
    The value for this cannot be higher than the Runway config file, it must
    be at the runway file itself or in a sub directory.

    Example:
        .. code-block:: yaml

            deployments:
            - modules:
                - path: my/local/module.cfn
                - my/local/module.cfn # same as above
                - ./ # module is in the root

    When the ``path`` is remote, Runway is responsible for fetching
    the resource and returning the location of it's cached path.
    The information for retrieving those sources can be controlled via
    runway rather than manually retrieving each one.

    Example:
        .. code-block:: yaml

            deployments:
            - modules:
                - path: git::git://github.com/your_handle/your_repo.git//my-module.cfn


    The ``path`` structure is based on the
    encoding found in
    `Terraform modules <https://www.terraform.io/docs/modules/sources.html>`_.

    The values parsed from the string are as follows:

    .. rubric:: source

    Determine if the source is local or remote. The initial
    prefix is used to determine this separated by `::` in the string.
    A path is considered local if it contains no source type value.

    Example:
        .. code-block:: yaml

            deployments:
                - modules:
                    # source is `git`
                    - path: git::git://github.com/foo/bar.git

    .. rubric:: uri

    The uniform resource identifier when targetting a remote resource.
    This instructs runway on where to retrieve your module.

    Example:
        .. code-block:: yaml

            deployments:
                - modules:
                    # uri is `git://github.com/foo/bar.git`
                    - path: git::git://github.com/foo/bar.git

    .. rubric:: location

    The relative location of the module files from the root
    directory. This value is specified as a path after the uri separated
    by `//`

    Example:
        .. code-block:: yaml

            deployments:
                - modules:
                    # location is `my/path`
                    - path: git::git://github.com/foo/bar.git//my/path

    .. rubric:: options

    The remaining options that are passed along to the
    Source. This is specified in the path following the `?` separator.
    Multiple option keys and values can be specified with the `&` as
    the separator. Each remote source can have different options for
    retrieval, please make sure to review individual source types
    to get more information on properly formatting.

    Example:
        .. code-block:: yaml

            deployments:
                - modules:
                    # options are `foo=bar&ba=bop`
                    - path: git::git://github.com/foo/bar.git//my/path?foo=bar&baz=bop

    """

    def __init__(self, module, env_root, cache_dir=None, git_source_class=Git):
        # type: (Union(str, Dict[str, str]), str, Optional[str], Optional[Git])-> Path
        """Path Configuration.

        Keyword Args:
            module (Union(str, Dict[str, str])): The module manifest or a string
                representation of a local path to a module.
            env_root (str):  The current environments root directory path
                string.
            cache_dir (Optional[str]): When a remote resource is requested it's
                Source object requires a cache directory to store it's request.
                This allows for an override of that default directory.
            git_source_class (Optional[Git]): Dependency injection for the `Git`
                type Source.

        References:
            `Git`_

        """
        if isinstance(module, six.string_types):
            module = {'path': module}  # type: Dict[str, str]

        self.git_source_class = git_source_class  # type: Git
        self.env_root = env_root  # type: str
        self.cache_dir = cache_dir  # type: str
        (self.source,
         self.uri,
         self.location,
         self.options) = self.parse(module)  # type: Tuple[str]
        self.module_root = self.__get_module_root_dir()  # type: str

    def __get_module_root_dir(self):
        # type: () -> str
        """Get module root directory.

        Retrieve the specific path location of the module. This can be static
        or dynamically generated by a remote resource Source object.
        """
        if self.location in ['.', '.' + os.sep]:
            return self.env_root
        if self.source != 'local':
            return self.__fetch_remote_source()
        return os.path.join(self.env_root, self.location)

    def __fetch_remote_source(self):
        # type: () -> Union(Git, None)
        """
        Switch based on the retrieved source of the path.

        Determine which remote Source type to fetch based on the source
        specified in the path variable.
        """
        if self.source == 'git':
            return self.git_source_class(**self.configuration).fetch()
        return None

    @property
    def configuration(self):
        # type: () -> Dict[str, str]
        """Transform object into configuration settings for remote Sources."""
        return {
            'source': self.source,
            'location': self.location,
            'uri': self.uri,
            'options': self.options,
            'cache_dir': self.cache_dir
        }

    @classmethod
    def parse(cls, module):
        # type: (Dict[str, str]) -> Tuple[str]
        """Retrieve the relevant elements of the path variable passed.

        Keyword Args:
            module (Dict[str, str]): The module manifest or a string
                representation of a local path to a module.

        Given a dictionary with a `path` parameter parse the value into
        it's specific components. The path structure is based on the
        encoding found in
        `Terraform modules <https://www.terraform.io/docs/modules/sources.html>`_.
        """
        source = 'local'  # type: str
        uri = ''  # type: str
        location = ''  # type: str
        options = ''  # type: str

        split_source_location = module.get('path', '').split('::')  # type: List[str]

        # Local path
        if len(split_source_location) != 2:
            location = split_source_location[0]  # type: str
            options = {}  # type: Dict
            return source, uri, location, options

        source = split_source_location[0]  # type: str
        temp_location = split_source_location[1]  # type: str

        location, options = cls.__parse_location_and_options(temp_location)  # type: List[str]
        uri, location = cls.__parse_uri_and_location(location)  # type: List[str]

        return source, uri, location, options

    @classmethod
    def __parse_uri_and_location(cls, uri_loc_str):
        # type: (str) -> List[str]
        """Given a string extract the uri and remaining location values.

        Keyword Args:
            uri_loc_str (str): The string that represents the uri and
                remaining location values.

        Separator used is `//`. It is expected the uri will contain a
        protocol reference, so if a remote uri is used those values
        will be concatenated together.
        """
        split_uri_location = uri_loc_str.split('//')  # type: List[str, str]
        location_string = ''  # type: str

        if len(split_uri_location) == 3:
            location_string = split_uri_location[2]  # type: str

        return [
            '//'.join([split_uri_location[0], split_uri_location[1]]),
            location_string
        ]

    @classmethod
    def __parse_location_and_options(cls, loc_opt_str):
        # type: (str) -> List[str]
        """Given a location string extract the location variable and the remote module options.

        Keyword Args:
            loc_opt_str (str): The string that represents the location
                and remaining option values

        Seperator used is `?`. Each of the options retrieved are then
        turned into a Dict for easy accessibility based on the `&` separator
        """
        split_location_options = loc_opt_str.split('?')  # type: List(str)
        location = split_location_options[0]  # type: str
        options = {}  # type: Dict

        if len(split_location_options) == 2:
            options = cls.__parse_options_dict(
                split_location_options[1]
            )  # type: Dict[str, str]

        return [location, options]

    @staticmethod
    def __parse_options_dict(options_str):
        # type: (str) -> Dict[str, str]
        """Convert the options string into a dict.

        Keyword Args:
            options_str (str): The options string that will
                be seperated into a dictionary based on the
                `&` separator
        """
        opts = options_str.split('&')  # type: List[str]
        res = {}  # Type: Dict

        for opt in opts:
            key, value = opt.split('=')  # type: List[str, str]
            res[key] = value  # type: str

        return res



"""CFNgin commands."""
import logging
import warnings

from runway.cfngin import __version__

from ... import session_cache
from ...config import render_parse_load as load_config
from ...context import Context
from ...providers.aws import default
from .base import BaseCommand
from .build import Build
from .destroy import Destroy
from .diff import Diff
from .graph import Graph
from .info import Info

LOGGER = logging.getLogger(__name__)


class Stacker(BaseCommand):
    """Stacker command."""

    name = "stacker"
    subcommands = (Build, Destroy, Info, Diff, Graph)

    DEPRECATION_MSG = ("Runway's Stacker CLI components have been deprecated "
                       "and will be removed in the next major release of "
                       "Runway.")

    def configure(self, options):
        """Configure CLI command."""
        warnings.warn(self.DEPRECATION_MSG, DeprecationWarning)
        LOGGER.warning(self.DEPRECATION_MSG)
        session_cache.default_profile = options.profile

        self.config = load_config(
            options.config.read(),
            environment=options.environment,
            validate=True,
        )

        options.provider_builder = default.ProviderBuilder(
            region=options.region,
            interactive=options.interactive,
            replacements_only=options.replacements_only,
            recreate_failed=options.recreate_failed,
            service_role=self.config.service_role,
        )

        options.context = Context(
            environment=options.environment,
            config=self.config,
            config_path=options.config.name,
            region=options.region,
            # Allow subcommands to provide any specific kwargs to the Context
            # that it wants.
            **options.get_context_kwargs(options)
        )

        super(Stacker, self).configure(options)
        if options.interactive:
            LOGGER.info("Using interactive AWS provider mode.")
        else:
            LOGGER.info("Using default AWS provider mode")

    def add_arguments(self, parser):
        """Add CLI arguments."""
        parser.add_argument("--version", action="version",
                            version="%%(prog)s %s" % (__version__,))



"""Import modules."""
from .cfngin import CFNgin  # noqa: F401

__all__ = ['CFNgin']

# added for stacker shim backward compatability.
# use of __version__ is deprecated and will be removed in 2.0.0.
__version__ = '1.7.0'



"""Tests for runway.cfngin.lookups.registry."""
import unittest

from mock import MagicMock

from runway.cfngin.exceptions import FailedVariableLookup, UnknownLookupType
from runway.cfngin.lookups.registry import CFNGIN_LOOKUP_HANDLERS
from runway.variables import Variable, VariableValueLookup

from ..factories import mock_context, mock_provider


class TestRegistry(unittest.TestCase):
    """Tests for runway.cfngin.lookups.registry."""

    def setUp(self):
        """Run before tests."""
        self.ctx = mock_context()
        self.provider = mock_provider()

    def test_autoloaded_lookup_handlers(self):
        """Test autoloaded lookup handlers."""
        handlers = [
            "output", "xref", "kms", "ssm", "ssmstore", "envvar", "rxref",
            "ami", "file", "split", "default", "hook_data", "dynamodb"
        ]
        for handler in handlers:
            try:
                CFNGIN_LOOKUP_HANDLERS[handler]
            except KeyError:
                assert False, 'Lookup handler: "{}" was not registered'.format(
                    handler
                )

    def test_resolve_lookups_string_unknown_lookup(self):
        """Test resolve lookups string unknown lookup."""
        with self.assertRaises(UnknownLookupType):
            Variable("MyVar", "${bad_lookup foo}")

    def test_resolve_lookups_list_unknown_lookup(self):
        """Test resolve lookups list unknown lookup."""
        with self.assertRaises(UnknownLookupType):
            Variable(
                "MyVar", [
                    "${bad_lookup foo}", "random string",
                ]
            )

    def resolve_lookups_with_output_handler_raise_valueerror(self, variable):
        """Resolve lookups with output handler raise valueerror.

        Mock output handler to throw ValueError, then run resolve_lookups
        on the given variable.

        """
        mock_handler = MagicMock(side_effect=ValueError("Error"))

        # find the only lookup in the variable
        for value in variable._value:  # pylint: disable=protected-access
            if isinstance(value, VariableValueLookup):
                value.handler = mock_handler

        with self.assertRaises(FailedVariableLookup) as result:
            variable.resolve(self.ctx, self.provider)

        self.assertIsInstance(result.exception.error, ValueError)

    def test_resolve_lookups_string_failed_variable_lookup(self):
        """Test resolve lookups string failed variable lookup."""
        variable = Variable("MyVar", "${output foo::bar}")
        self.resolve_lookups_with_output_handler_raise_valueerror(variable)

    def test_resolve_lookups_list_failed_variable_lookup(self):
        """Test resolve lookups list failed variable lookup."""
        variable = Variable(
            "MyVar", [
                "random string", "${output foo::bar}", "random string",
            ]
        )
        self.resolve_lookups_with_output_handler_raise_valueerror(variable)



"""Callback URL Retriever.

Dependency pre-hook responsible for ensuring correct
callback urls are retrieved or a temporary one is used in it's place.
"""

import logging

from typing import Any, Dict, Optional  # pylint: disable=unused-import

from runway.cfngin.providers.base import BaseProvider  # pylint: disable=unused-import
from runway.cfngin.context import Context  # noqa pylint: disable=unused-import
from runway.cfngin.session_cache import get_session

LOGGER = logging.getLogger(__name__)


def get(context,  # pylint: disable=unused-argument
        provider,
        **kwargs
       ):  # noqa: E124
    # type: (Context, BaseProvider, Optional[Dict[str, Any]]) -> Dict
    """Retrieve the callback URLs for User Pool Client Creation.

    When the User Pool is created a Callback URL is required. During a post
    hook entitled ``client_updater`` these Callback URLs are updated to that
    of the Distribution. Before then we need to ensure that if a Client
    already exists that the URLs for that client are used to prevent any
    interuption of service during deploy.

    Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance

    Keyword Args:
        user_pool_id (str): The ID of the User Pool to check for a client
        stack_name (str) The name of the stack to check against
    """
    session = get_session(provider.region)
    cloudformation_client = session.client('cloudformation')
    cognito_client = session.client('cognito-idp')

    context_dict = {}
    context_dict['callback_urls'] = ['https://example.tmp']

    try:
        # Return the current stack if one exists
        stack_desc = cloudformation_client.describe_stacks(
            StackName=kwargs['stack_name']
        )
        # Get the client_id from the outputs
        outputs = stack_desc['Stacks'][0]['Outputs']

        if kwargs['user_pool_arn']:
            user_pool_id = kwargs['user_pool_arn'].split('/')[-1:][0]
        else:
            user_pool_id = [
                o['OutputValue'] for o in outputs if o['OutputKey'] == 'AuthAtEdgeUserPoolId'
            ][0]

        client_id = [o['OutputValue'] for o in outputs if o['OutputKey'] == 'AuthAtEdgeClient'][0]

        # Poll the user pool client information
        resp = cognito_client.describe_user_pool_client(
            UserPoolId=user_pool_id,
            ClientId=client_id
        )

        # Retrieve the callbacks
        callbacks = resp['UserPoolClient']['CallbackURLs']

        if callbacks:
            context_dict['callback_urls'] = callbacks
        return context_dict
    except Exception:  # pylint: disable=broad-except
        return context_dict



"""Blueprint for the Authorization@Edge implementation of a Static Site.

Described in detail in this blogpost:
https://aws.amazon.com/blogs/networking-and-content-delivery/authorizationedge-how-to-use-lambdaedge-and-json-web-tokens-to-enhance-web-application-security/
"""

import logging
from typing import Any, Dict, List  # pylint: disable=unused-import

import awacs.s3
import awacs.logs
from awacs.aws import Allow, Principal, Statement
from troposphere import (AccountId, Join, NoValue, StackName,  # noqa pylint: disable=unused-import
                         awslambda, cloudfront, iam, s3)

from .staticsite import StaticSite

LOGGER = logging.getLogger('runway')


class AuthAtEdge(StaticSite):
    """Auth@Edge Blueprint."""

    IAM_ARN_PREFIX = 'arn:aws:iam::aws:policy/service-role/'

    VARIABLES = {
        'AcmCertificateArn': {'type': str,
                              'default': '',
                              'description': '(Optional) Cert ARN for site'},
        'Aliases': {'type': list,
                    'default': [],
                    'description': '(Optional) Domain aliases the '
                                   'distribution'},
        'LogBucketName': {'type': str,
                          'default': '',
                          'description': 'S3 bucket for CF logs'},
        'OAuthScopes': {'type': list,
                        'default': [],
                        'description': 'OAuth2 Scopes'},
        'DisableCloudFront': {'type': bool,
                              'default': False,
                              'description': 'Whether to disable CF'},
        'PriceClass': {'type': str,
                       'default': 'PriceClass_100',  # US/Europe
                       'description': 'CF price class for the distribution.'},
        'RedirectPathSignIn': {'type': str,
                               'default': '/parseauth',
                               'description': 'Auth@Edge: The URL that should '
                                              'handle the redirect from Cognito '
                                              'after sign-in.'},
        'RedirectPathAuthRefresh': {'type': str,
                                    'default': '/refreshauth',
                                    'description': 'The URL path that should '
                                                   'handle the JWT refresh request.'},
        'RewriteDirectoryIndex': {'type': str,
                                  'default': '',
                                  'description': 'The path for rewriting the directory index'},
        'RoleBoundaryArn': {'type': str,
                            'default': '',
                            'description': '(Optional) IAM Role Boundary to '
                                           'apply to any created IAM Roles.'},
        'NonSPAMode': {'type': bool,
                       'default': False,
                       'description': 'Whether Auth@Edge should omit SPA specific settings'},
        'SignOutUrl': {'type': str,
                       'default': '/signout',
                       'description': 'The URL path that you can visit to sign-out.'},
        'WAFWebACL': {'type': str,
                      'default': '',
                      'description': '(Optional) WAF id to associate with the '
                                     'distribution.'},
        'custom_error_responses': {'type': list,
                                   'default': [],
                                   'description': '(Optional) Custom error '
                                                  'responses.'},
    }

    def create_template(self):
        # type: () -> None
        """Create the Blueprinted template for Auth@Edge."""
        self.template.set_version('2010-09-09')
        self.template.set_description(
            'Authorization@Edge Static Website - Bucket, Lambdas, and Distribution'
        )

        # Resources
        bucket = self.add_bucket()
        # self.add_bucket_policy(bucket)
        oai = self.add_origin_access_identity()
        bucket_policy = self.add_cloudfront_bucket_policy(bucket, oai)
        # @TODO: Make this available in Auth@Edge
        lambda_function_associations = []

        if self.directory_index_specified:
            index_rewrite = self._get_index_rewrite_role_function_and_version()
            lambda_function_associations = self.get_directory_index_lambda_association(
                lambda_function_associations,
                index_rewrite['version']
            )

        # Auth@Edge Lambdas
        check_auth_name = 'CheckAuth'
        check_auth_lambda = self.get_auth_at_edge_lambda_and_ver(
            check_auth_name,
            'Check Authorization information for request',
            'check_auth',
            self.add_lambda_execution_role(
                'CheckAuthLambdaExecutionRole',
                check_auth_name
            )
        )
        http_headers_name = 'HttpHeaders'
        http_headers_lambda = self.get_auth_at_edge_lambda_and_ver(
            http_headers_name,
            'Additional Headers added to every response',
            'http_headers',
            self.add_lambda_execution_role(
                'HttpHeadersLambdaExecutionRole',
                http_headers_name
            )
        )
        parse_auth_name = 'ParseAuth'
        parse_auth_lambda = self.get_auth_at_edge_lambda_and_ver(
            parse_auth_name,
            'Parse the Authorization Headers/Cookies for the request',
            'parse_auth',
            self.add_lambda_execution_role(
                'ParseAuthLambdaExecutionRole',
                parse_auth_name
            )
        )
        refresh_auth_name = 'RefreshAuth'
        refresh_auth_lambda = self.get_auth_at_edge_lambda_and_ver(
            refresh_auth_name,
            'Refresh the Authorization information when expired',
            'refresh_auth',
            self.add_lambda_execution_role(
                'RefreshAuthLambdaExecutionRole',
                refresh_auth_name
            )
        )
        sign_out_name = 'SignOut'
        sign_out_lambda = self.get_auth_at_edge_lambda_and_ver(
            sign_out_name,
            'Sign the User out of the application',
            'sign_out',
            self.add_lambda_execution_role(
                'SignOutLambdaExecutionRole',
                sign_out_name
            )
        )

        # CloudFront Distribution
        distribution_options = self.get_distribution_options(
            bucket,
            oai,
            lambda_function_associations,
            check_auth_lambda['version'],
            http_headers_lambda['version'],
            parse_auth_lambda['version'],
            refresh_auth_lambda['version'],
            sign_out_lambda['version']
        )
        self.add_cloudfront_distribution(
            bucket_policy,
            distribution_options
        )

    def get_auth_at_edge_lambda_and_ver(self,
                                        title,  # type: str
                                        description,  # type: str
                                        handle,  # type: str
                                        role  # iam.Role
                                       ):  # noqa: E124
        # type: (...) -> Dict[str, Any]
        """Create a lambda function and its version.

        Args:
            title (str): The name of the function in PascalCase
            description (str): Description to be displayed in the
                lambda panel
            handle (str): The underscore separated representation
                of the name of the lambda. This handle is used to
                determine the handler for the lambda as well as
                identify the correct Code hook_data information.
            role (IAM.Role): The Lambda Execution Role
        """
        function = self.get_auth_at_edge_lambda(
            title,
            description,
            handle,
            role
        )

        return {
            'function': function,
            'version': self.add_version(title, function)
        }

    def get_auth_at_edge_lambda(self,
                                title,   # type: str
                                description,   # type: str
                                handler,  # type: str
                                role  # type: iam.Role
                               ):  # noqa: E124
        # type: (...) -> awslambda.Function
        """Create an Auth@Edge lambda resource.

        Args:
            name (str): The name of the function in PascalCase
            description (str): Description to be displayed in the
                lambda panel
            handle (str): The underscore separated representation
                of the name of the lambda. This handle is used to
                determine the handler for the lambda as well as
                identify the correct Code hook_data information.
            role (IAM.Role): The Lambda Execution Role
        """
        return self.template.add_resource(
            awslambda.Function(
                title,
                Code=self.context.hook_data['aae_lambda_config'][handler],
                Description=description,
                Handler='__init__.handler',
                Role=role.get_att("Arn"),
                Runtime='python3.7'
            )
        )

    def add_version(self,
                    title,  # type: str
                    lambda_function  # type: awslambda.Function
                   ):  # noqa E214
        # type: (...) -> awslambda.Version
        """Create a version association with a Lambda@Edge function.

        In order to ensure different versions of the function
        are appropriately uploaded a hash based on the code of the
        lambda is appended to the name. As the code changes so
        will this hash value.

        Args:
            title (str): The name of the function in PascalCase
            role (awslambda.Function): The Lambda function
        """
        s3_key = lambda_function.properties['Code'].to_dict()['S3Key']
        code_hash = s3_key.split('.')[0].split('-')[-1]
        return self.template.add_resource(
            awslambda.Version(
                title + 'Ver' + code_hash,
                FunctionName=lambda_function.ref()
            )
        )

    def get_distribution_options(self,
                                 bucket,  # type: s3.Bucket
                                 oai,  # type: cloudfront.CloudFrontOriginAccessIdentity
                                 lambda_funcs,  # type: List[cloudfront.LambdaFunctionAssociation]
                                 check_auth_lambda_version,  # type: awslambda.Version
                                 http_headers_lambda_version,  # type: awslambda.Version
                                 parse_auth_lambda_version,  # type: awslambda.Version
                                 refresh_auth_lambda_version,  # type: awslambda.Version
                                 sign_out_lambda_version  # type: awslambda.Version
                                ):  # noqa: E124
        # type: (...) -> Dict[str, Any]
        """Retrieve the options for our CloudFront distribution.

        Keyword Args:
            bucket (dict): The bucket resource
            oai (dict): The origin access identity resource

        Return:
            dict: The CloudFront Distribution Options

        """
        variables = self.get_variables()

        default_cache_behavior_lambdas = lambda_funcs
        default_cache_behavior_lambdas.append(
            cloudfront.LambdaFunctionAssociation(
                EventType='viewer-request',
                LambdaFunctionARN=check_auth_lambda_version.ref()
            )
        )
        default_cache_behavior_lambdas.append(
            cloudfront.LambdaFunctionAssociation(
                EventType='origin-response',
                LambdaFunctionARN=http_headers_lambda_version.ref()
            )
        )

        return {
            'Aliases': self.add_aliases(),
            'Origins': [
                cloudfront.Origin(
                    DomainName=Join(
                        '.',
                        [bucket.ref(),
                         's3.amazonaws.com']),
                    S3OriginConfig=cloudfront.S3OriginConfig(
                        OriginAccessIdentity=Join(
                            '',
                            ['origin-access-identity/cloudfront/',
                             oai.ref()])
                    ),
                    Id='protected-bucket'
                )
            ],
            'CacheBehaviors': [
                cloudfront.CacheBehavior(
                    PathPattern=variables['RedirectPathSignIn'],
                    Compress=True,
                    ForwardedValues=cloudfront.ForwardedValues(
                        QueryString=True
                    ),
                    LambdaFunctionAssociations=[
                        cloudfront.LambdaFunctionAssociation(
                            EventType='viewer-request',
                            LambdaFunctionARN=parse_auth_lambda_version.ref()
                        )
                    ],
                    TargetOriginId='protected-bucket',
                    ViewerProtocolPolicy="redirect-to-https"
                ),
                cloudfront.CacheBehavior(
                    PathPattern=variables['RedirectPathAuthRefresh'],
                    Compress=True,
                    ForwardedValues=cloudfront.ForwardedValues(
                        QueryString=True
                    ),
                    LambdaFunctionAssociations=[
                        cloudfront.LambdaFunctionAssociation(
                            EventType='viewer-request',
                            LambdaFunctionARN=refresh_auth_lambda_version.ref()
                        )
                    ],
                    TargetOriginId='protected-bucket',
                    ViewerProtocolPolicy="redirect-to-https"
                ),
                cloudfront.CacheBehavior(
                    PathPattern=variables['SignOutUrl'],
                    Compress=True,
                    ForwardedValues=cloudfront.ForwardedValues(
                        QueryString=True
                    ),
                    LambdaFunctionAssociations=[
                        cloudfront.LambdaFunctionAssociation(
                            EventType='viewer-request',
                            LambdaFunctionARN=sign_out_lambda_version.ref()
                        )
                    ],
                    TargetOriginId='protected-bucket',
                    ViewerProtocolPolicy="redirect-to-https"
                ),
            ],
            'DefaultCacheBehavior': cloudfront.DefaultCacheBehavior(
                AllowedMethods=['GET', 'HEAD'],
                Compress=True,
                DefaultTTL='86400',
                ForwardedValues=cloudfront.ForwardedValues(
                    QueryString=True,
                ),
                LambdaFunctionAssociations=default_cache_behavior_lambdas,
                TargetOriginId='protected-bucket',
                ViewerProtocolPolicy='redirect-to-https'
            ),
            'DefaultRootObject': 'index.html',
            'Logging': self.add_logging_bucket(),
            'PriceClass': variables['PriceClass'],
            'Enabled': True,
            'WebACLId': self.add_web_acl(),
            'CustomErrorResponses': self._get_error_responses(),
            'ViewerCertificate': self.add_acm_cert()
        }

    def _get_error_responses(self):
        """Return error response based on site stack variables.

        When custom_error_responses are defined return those, if running
        in NonSPAMode return nothing, or return the standard error responses
        for a SPA.
        """
        variables = self.get_variables()
        if variables['custom_error_responses']:
            return [
                cloudfront.CustomErrorResponse(
                    ErrorCode=response['ErrorCode'],
                    ResponseCode=response['ResponseCode'],
                    ResponsePagePath=response['ResponsePagePath']
                ) for response in variables['custom_error_responses']
            ]
        if variables['NonSPAMode']:
            return []
        return [
            cloudfront.CustomErrorResponse(
                ErrorCode=404,
                ResponseCode=200,
                ResponsePagePath='/index.html'
            )
        ]

    def _get_cloudfront_bucket_policy_statements(self, bucket, oai):
        return [
            Statement(
                Action=[awacs.s3.GetObject],
                Effect=Allow,
                Principal=Principal(
                    'CanonicalUser',
                    oai.get_att('S3CanonicalUserId')
                ),
                Resource=[
                    Join('', [bucket.get_att('Arn'), '/*'])
                ]
            ),
            Statement(
                Action=[awacs.s3.ListBucket],
                Effect=Allow,
                Principal=Principal(
                    'CanonicalUser',
                    oai.get_att('S3CanonicalUserId')
                ),
                Resource=[bucket.get_att('Arn')]
            )
        ]



"""Cloudformation module."""
import logging

from ..cfngin import CFNgin
from . import RunwayModule

LOGGER = logging.getLogger('runway')


class CloudFormation(RunwayModule):
    """CloudFormation (Stacker) Runway Module."""

    def deploy(self):
        """Run stacker build."""
        cfngin = CFNgin(self.context,
                        parameters=self.options['parameters'],
                        sys_path=self.path)
        cfngin.deploy(force=self.options['environment'])

    def destroy(self):
        """Run stacker destroy."""
        cfngin = CFNgin(self.context,
                        parameters=self.options['parameters'],
                        sys_path=self.path)
        cfngin.destroy(force=self.options['environment'])

    def plan(self):
        """Run stacker diff."""
        cfngin = CFNgin(self.context,
                        parameters=self.options['parameters'],
                        sys_path=self.path)
        cfngin.plan(force=self.options['environment'])



"""Re-usable class for Serverless testing."""
import os
import tempfile
import boto3

from runway.s3_util import (download,
                            get_matching_s3_keys,
                            purge_and_delete_bucket,
                            purge_bucket)
from runway.util import change_dir, find_cfn_output
from runway.module.serverless import get_src_hash
from runway.commands.modules_command import assume_role
from runway.hooks.staticsite.util import get_hash_of_files

from integration_tests.test_serverless.test_serverless import Serverless
from integration_tests.util import run_command


class ServerlessTest(Serverless):
    """Class for Serverless tests."""

    ENVS = ('dev', 'test')

    def __init__(self, template, templates_dir, environment, logger):
        """Initialize class."""
        self.template_name = template
        self.templates_dir = templates_dir
        self.template_dir = os.path.join(templates_dir, template)
        self.environment = environment
        self.logger = logger
        self.runway_config_path = os.path.join(templates_dir, template, 'runway.yml')

    def get_serverless_bucket(self, stack_name, session):
        """Get bucket created by serverless."""
        cfn_client = session.client('cloudformation')
        stacks = cfn_client.describe_stacks(StackName=stack_name)
        self.logger.debug('Found Stacks: ', stacks)
        return find_cfn_output('ServerlessDeploymentBucketName', stacks['Stacks'][0]['Outputs'])

    def get_promote_zip_bucket(self, runway_config):
        """Get promotezip bucket from runway config."""
        deployment = runway_config['deployments'][0]
        module = deployment.get('modules')[0]
        return module.get('options', {})\
                     .get('promotezip', {})\
                     .get('bucketname')

    def get_session(self, role_arn):
        """Get assumed role session."""
        self.logger.info('Assuming role: %s', role_arn)
        env_vars = os.environ.copy()
        role = assume_role(role_arn, env_vars=env_vars)
        return boto3.session.Session(role['AWS_ACCESS_KEY_ID'],
                                     role['AWS_SECRET_ACCESS_KEY'],
                                     role['AWS_SESSION_TOKEN'],
                                     'us-east-1')

    def get_configs(self):
        """Get Runway and Serverless parsed configs."""
        return {'Runway': self.parse_config(os.path.join(self.template_dir, 'runway.yml')),
                'Serverless':
                self.parse_config(os.path.join(self.template_dir, 'serverless.yml'))}

    def get_zip_hashes(self, env):
        """Get serverless and runway buckets."""
        configs = self.get_configs()
        promotezip_bucket = self.get_promote_zip_bucket(configs['Runway'])

        session = self.get_session(configs['Runway']['deployments'][0].get('assume_role')[env])
        sls_bucket = self.get_serverless_bucket('-'.join([configs['Serverless'].get('service'),
                                                          env]),
                                                session)

        hashes = get_src_hash(configs['Serverless'], self.template_dir)
        zip_dirs = {'Runway': tempfile.mkdtemp(), 'Serverless': tempfile.mkdtemp()}
        for key, value in hashes.items():
            # don't need to pass session for promotezip bucket as it will use default
            # creds when downloading
            hash_zip = value + '.zip'
            func_zip = os.path.basename(key) + '.zip'
            self.logger.info('Download zip %s from %s', promotezip_bucket, hash_zip)
            download(promotezip_bucket, hash_zip, os.path.join(zip_dirs['Runway'], func_zip))

            for s3key in get_matching_s3_keys(sls_bucket,
                                              'serverless/{0}/{1}'.format(configs['Serverless'].get('service'), env), # noqa pylint: disable=line-too-long
                                              os.path.basename(key) + ".zip",
                                              session):
                self.logger.info('Downloading zip %s from %s', s3key, sls_bucket)
                # use session for this call as the bucket is in a different account
                download(sls_bucket, s3key,
                         os.path.join(zip_dirs['Serverless'], func_zip),
                         session)

        return [get_hash_of_files(value) for key, value in zip_dirs.items()]

    def run_runway(self, template, command='deploy'):
        """Deploy serverless template."""
        template_dir = os.path.join(self.templates_dir, template)
        if os.path.isdir(template_dir):
            self.logger.info('Executing test "%s" in directory "%s"', template, template_dir)
            with change_dir(template_dir):
                self.logger.info('Running "runway %s" on %s', command, template_dir)
                return run_command(['runway', command], self.environment)
        else:
            self.logger.error('Directory not found: %s', template_dir)
            return 1

    def run(self):
        """Run tests."""
        self.clean()
        hashes = []
        for env in self.ENVS:
            self.set_environment(env)
            assert self.run_runway(self.template_name) == 0,\
                '{}: Failed to deploy in {} environment'.format(self.template_name, env)
            hashes.append(self.get_zip_hashes(env))

        self.logger.info("Verifying zips are the same")
        assert hashes[1:] == hashes[:-1],\
            '{}: Hash sums do not match for all zipped functions'.format(self.template_name)

    def teardown(self):
        """Teardown any created resources."""
        self.logger.info('Tearing down: %s', self.template_name)
        configs = self.get_configs()

        for env in self.ENVS:
            self.set_environment(env)
            session = self.get_session(configs['Runway']['deployments'][0]
                                       .get('assume_role')[env])
            sls_bucket = self.get_serverless_bucket('-'.join([configs['Serverless']
                                                              .get('service'),
                                                              env]),
                                                    session)
            purge_bucket(sls_bucket, session=session)
            self.run_runway(self.template_name, 'destroy')

        purge_and_delete_bucket(self.get_promote_zip_bucket(configs['Runway']))
        self.clean()



"""Tests for runway.cfngin.config."""
# pylint: disable=no-member
import sys
import unittest

from yaml.constructor import ConstructorError

from runway.cfngin import exceptions
from runway.cfngin.config import (Config, Stack, dump, load, parse,
                                  process_remote_sources, render,
                                  render_parse_load)
from runway.cfngin.environment import parse_environment
from runway.cfngin.lookups.registry import CFNGIN_LOOKUP_HANDLERS

CONFIG = """a: $a
b: $b
c: $c"""


class TestConfig(unittest.TestCase):
    """Tests for runway.cfngin.config."""

    def test_render_missing_env(self):
        """Test render missing env."""
        env = {"a": "A"}
        with self.assertRaises(exceptions.MissingEnvironment) as expected:
            render(CONFIG, env)
        self.assertEqual(expected.exception.key, "b")

    def test_render_no_variable_config(self):
        """Test render no variable config."""
        config = render("namespace: prod", {})
        self.assertEqual("namespace: prod", config)

    def test_render_valid_env_substitution(self):
        """Test render valid env substitution."""
        config = render("namespace: $namespace", {"namespace": "prod"})
        self.assertEqual("namespace: prod", config)

    def test_render_blank_env_values(self):
        """Test ender blank env values."""
        conf = """namespace: ${namespace}"""
        env = parse_environment("""namespace:""")
        config = render(conf, env)
        self.assertEqual("namespace: ", config)
        env = parse_environment("""namespace: !!str""")
        config = render(conf, env)
        self.assertEqual("namespace: !!str", config)

    def test_config_validate_missing_stack_source(self):
        """Test config validate missing stack source."""
        config = Config({
            "namespace": "prod",
            "stacks": [
                {
                    "name": "bastion"}]})
        with self.assertRaises(exceptions.InvalidConfig) as ex:
            config.validate()

        stack_errors = ex.exception.errors['stacks'][0]
        assert stack_errors['template_path'][0].__str__() == \
            "class_path or template_path is required."
        assert stack_errors['class_path'][0].__str__() == \
            "class_path or template_path is required."

    def test_config_validate_missing_stack_source_when_locked(self):
        """Test config validate missing stack source when locked."""
        config = Config({
            "namespace": "prod",
            "stacks": [
                {
                    "name": "bastion",
                    "locked": True}]})
        config.validate()

    def test_config_validate_stack_class_and_template_paths(self):
        """Test config validate stack class and template paths."""
        config = Config({
            "namespace": "prod",
            "stacks": [
                {
                    "name": "bastion",
                    "class_path": "foo",
                    "template_path": "bar"}]})
        with self.assertRaises(exceptions.InvalidConfig) as ex:
            config.validate()

        stack_errors = ex.exception.errors['stacks'][0]
        assert stack_errors['template_path'][0].__str__() == \
            "class_path cannot be present when template_path is provided."
        assert stack_errors['class_path'][0].__str__() == \
            "template_path cannot be present when class_path is provided."

    def test_config_validate_missing_name(self):
        """Test config validate missing name."""
        config = Config({
            "namespace": "prod",
            "stacks": [
                {
                    "class_path": "blueprints.Bastion"}]})
        with self.assertRaises(exceptions.InvalidConfig) as ex:
            config.validate()

        error = ex.exception.errors['stacks'][0]['name'].errors[0]
        assert error.__str__() == "This field is required."

    def test_config_validate_duplicate_stack_names(self):
        """Test config validate duplicate stack names."""
        config = Config({
            "namespace": "prod",
            "stacks": [
                {
                    "name": "bastion",
                    "class_path": "blueprints.Bastion"},
                {
                    "name": "bastion",
                    "class_path": "blueprints.BastionV2"}]})
        with self.assertRaises(exceptions.InvalidConfig) as ex:
            config.validate()

        error = ex.exception.errors['stacks'][0]
        assert error.__str__() == "Duplicate stack bastion found at index 0."

    def test_dump_unicode(self):
        """Test dump unicode."""
        config = Config()
        config.namespace = "test"
        assert dump(config) == b"""namespace: test
stacks: []
"""

        config = Config({"namespace": "test"})
        # Ensure that we're producing standard yaml, that doesn't include
        # python specific objects.
        assert dump(config) != b"namespace: !!python/unicode 'test'\n"
        assert dump(config) == b"""namespace: test
stacks: []
"""

    def test_parse_tags(self):
        """Test parse tags."""
        config = parse("""
        namespace: prod
        tags:
          "a:b": "c"
          "hello": 1
          simple_tag: simple value
        """)
        assert config.tags == {
            "a:b": "c",
            "hello": "1",
            "simple_tag": "simple value"
        }

    def test_parse_with_arbitrary_anchors(self):
        """Test parse with arbitrary anchors."""
        config = parse("""
        namespace: prod
        common_variables: &common_variables
          Foo: bar
        stacks:
        - name: vpc
          class_path: blueprints.VPC
          variables:
            << : *common_variables
        """)

        stack = config.stacks[0]
        assert stack.variables == {"Foo": "bar"}

    def test_parse_with_deprecated_parameters(self):
        """Test parse with deprecated parameters."""
        config = parse("""
        namespace: prod
        stacks:
        - name: vpc
          class_path: blueprints.VPC
          parameters:
            Foo: bar
        """)
        with self.assertRaises(exceptions.InvalidConfig) as ex:
            config.validate()

        error = ex.exception.errors['stacks'][0]['parameters'][0]
        assert error.__str__() == (
            "DEPRECATION: Stack definition vpc contains deprecated 'parameters'"
            ", rather than 'variables'. You are required to update your config. "
            "See https://docs.onica.com/projects/runway/en/release/cfngin/config.html#variables "
            "for additional information."
        )

    def test_config_build(self):
        """Test config build."""
        vpc = Stack({"name": "vpc", "class_path": "blueprints.VPC"})
        config = Config({"namespace": "prod", "stacks": [vpc]})
        assert config.namespace == 'prod'
        assert config.stacks[0].name == 'vpc'
        assert config["namespace"] == 'prod'
        config.validate()

    def test_parse(self):
        """Test parse."""
        config_with_lists = """
        namespace: prod
        stacker_bucket: cfngin-prod
        pre_build:
          - path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        post_build:
          - path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        pre_destroy:
          - path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        post_destroy:
          - path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        package_sources:
          s3:
            - bucket: acmecorpbucket
              key: public/acmecorp-blueprints-v1.zip
            - bucket: examplecorpbucket
              key: public/examplecorp-blueprints-v2.tar.gz
              requester_pays: true
            - bucket: anotherexamplebucket
              key: example-blueprints-v3.tar.gz
              use_latest: false
              paths:
                - foo
              configs:
                - foo/config.yml
          git:
            - uri: git@github.com:acmecorp/stacker_blueprints.git
            - uri: git@github.com:remind101/stacker_blueprints.git
              tag: 1.0.0
              paths:
                - stacker_blueprints
            - uri: git@github.com:contoso/webapp.git
              branch: staging
            - uri: git@github.com:contoso/foo.git
              commit: 12345678
              paths:
                - bar
              configs:
                - bar/moreconfig.yml
        tags:
          environment: production
        stacks:
        - name: vpc
          class_path: blueprints.VPC
          variables:
            PrivateSubnets:
            - 10.0.0.0/24
        - name: bastion
          class_path: blueprints.Bastion
          requires: ['vpc']
          variables:
            VpcId: ${output vpc::VpcId}
        """
        config_with_dicts = """
        namespace: prod
        stacker_bucket: cfngin-prod
        pre_build:
          prebuild_createdomain:
            path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        post_build:
          postbuild_createdomain:
            path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        pre_destroy:
          predestroy_createdomain:
            path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        post_destroy:
          postdestroy_createdomain:
            path: runway.cfngin.hooks.route53.create_domain
            required: true
            enabled: true
            args:
              domain: mydomain.com
        package_sources:
          s3:
            - bucket: acmecorpbucket
              key: public/acmecorp-blueprints-v1.zip
            - bucket: examplecorpbucket
              key: public/examplecorp-blueprints-v2.tar.gz
              requester_pays: true
            - bucket: anotherexamplebucket
              key: example-blueprints-v3.tar.gz
              use_latest: false
              paths:
                - foo
              configs:
                - foo/config.yml
          git:
            - uri: git@github.com:acmecorp/stacker_blueprints.git
            - uri: git@github.com:remind101/stacker_blueprints.git
              tag: 1.0.0
              paths:
                - stacker_blueprints
            - uri: git@github.com:contoso/webapp.git
              branch: staging
            - uri: git@github.com:contoso/foo.git
              commit: 12345678
              paths:
                - bar
              configs:
                - bar/moreconfig.yml
        tags:
          environment: production
        stacks:
          vpc:
            class_path: blueprints.VPC
            variables:
              PrivateSubnets:
              - 10.0.0.0/24
          bastion:
            class_path: blueprints.Bastion
            requires: ['vpc']
            variables:
              VpcId: ${output vpc::VpcId}
        """

        for raw_config in [config_with_lists, config_with_dicts]:
            config = parse(raw_config)

            config.validate()

            assert config.namespace == 'prod'
            assert config.stacker_bucket == 'cfngin-prod'

            for hooks in [config.pre_build, config.post_build,
                          config.pre_destroy, config.post_destroy]:
                assert hooks[0].path == 'runway.cfngin.hooks.route53.create_domain'
                assert hooks[0].required
                assert hooks[0].args == {"domain": "mydomain.com"}

            assert config.package_sources.s3[0].bucket == 'acmecorpbucket'
            assert config.package_sources.s3[0].key == \
                'public/acmecorp-blueprints-v1.zip'
            assert config.package_sources.s3[1].bucket == 'examplecorpbucket'
            assert config.package_sources.s3[1].key == \
                'public/examplecorp-blueprints-v2.tar.gz'
            assert config.package_sources.s3[1].requester_pays
            assert not config.package_sources.s3[2].use_latest

            assert config.package_sources.git[0].uri == \
                'git@github.com:acmecorp/stacker_blueprints.git'
            assert config.package_sources.git[1].uri == \
                'git@github.com:remind101/stacker_blueprints.git'
            assert config.package_sources.git[1].tag == '1.0.0'
            assert config.package_sources.git[1].paths == ['stacker_blueprints']
            assert config.package_sources.git[2].branch == 'staging'

            assert config.tags == {"environment": "production"}

            assert len(config.stacks) == 2
            self.assertEqual(len(config.stacks), 2)

            vpc_index = next(
                i for (i, d) in enumerate(config.stacks) if d.name == 'vpc'
            )
            vpc = config.stacks[vpc_index]
            self.assertEqual(vpc.name, "vpc")
            self.assertEqual(vpc.class_path, "blueprints.VPC")
            self.assertEqual(vpc.requires, None)
            self.assertEqual(vpc.variables,
                             {"PrivateSubnets": ["10.0.0.0/24"]})

            bastion_index = next(
                i for (i, d) in enumerate(config.stacks) if d.name == "bastion"
            )
            bastion = config.stacks[bastion_index]
            self.assertEqual(bastion.name, "bastion")
            self.assertEqual(bastion.class_path, "blueprints.Bastion")
            self.assertEqual(bastion.requires, ["vpc"])
            self.assertEqual(bastion.variables,
                             {"VpcId": "${output vpc::VpcId}"})

    def test_dump_complex(self):
        """Test dump complex."""
        config = Config({
            "namespace": "prod",
            "stacks": [
                Stack({
                    "name": "vpc",
                    "class_path": "blueprints.VPC"}),
                Stack({
                    "name": "bastion",
                    "class_path": "blueprints.Bastion",
                    "requires": ["vpc"]})]})

        self.assertEqual(dump(config), b"""namespace: prod
stacks:
- class_path: blueprints.VPC
  enabled: true
  locked: false
  name: vpc
  protected: false
  termination_protection: false
- class_path: blueprints.Bastion
  enabled: true
  locked: false
  name: bastion
  protected: false
  requires:
  - vpc
  termination_protection: false
""")

    def test_load_register_custom_lookups(self):
        """Test load register custom lookups."""
        config = Config({
            "lookups": {
                "custom": "importlib.import_module"}})
        load(config)
        self.assertTrue(callable(CFNGIN_LOOKUP_HANDLERS["custom"]))

    def test_load_adds_sys_path(self):
        """Test load adds sys path."""
        config = Config({"sys_path": "/foo/bar"})
        load(config)
        self.assertIn("/foo/bar", sys.path)

    def test_process_empty_remote_sources(self):
        """Test process empty remote sources."""
        config = """
        namespace: prod
        stacks:
          - name: vpc
            class_path: blueprints.VPC
        """
        self.assertEqual(config, process_remote_sources(config))

    def test_lookup_with_sys_path(self):
        """Test lookup with sys path."""
        config = Config({
            "sys_path": "./tests/cfngin",
            "lookups": {
                "custom": "fixtures.mock_lookups.handler"}})
        load(config)
        self.assertTrue(callable(CFNGIN_LOOKUP_HANDLERS["custom"]))

    def test_render_parse_load_namespace_fallback(self):
        """Test render parse load namespace fallback."""
        conf = """
        stacks:
        - name: vpc
          class_path: blueprints.VPC
        """
        config = render_parse_load(
            conf, environment={"namespace": "prod"}, validate=False)
        config.validate()
        assert config.namespace == 'prod'

    def test_allow_most_keys_to_be_duplicates_for_overrides(self):
        """Test allow most keys to be duplicates for overrides."""
        yaml_config = """
        namespace: prod
        stacks:
          - name: vpc
            class_path: blueprints.VPC
            variables:
              CIDR: 192.168.1.0/24
              CIDR: 192.168.2.0/24
        """
        doc = parse(yaml_config)
        print(doc.to_primitive())
        self.assertEqual(
            doc["stacks"][0]["variables"]["CIDR"], "192.168.2.0/24"
        )
        yaml_config = """
        default_variables: &default_variables
          CIDR: 192.168.1.0/24
        namespace: prod
        stacks:
          - name: vpc
            class_path: blueprints.VPC
            variables:
              << : *default_variables
              CIDR: 192.168.2.0/24
        """
        doc = parse(yaml_config)
        self.assertEqual(
            doc["stacks"][0]["variables"]["CIDR"], "192.168.2.0/24"
        )

    def test_raise_constructor_error_on_keyword_duplicate_key(self):
        """Some keys should never have a duplicate sibling.

        For example we treat `class_path` as a special "keyword" and
        disallow dupes.

        """
        yaml_config = """
        namespace: prod
        stacks:
          - name: vpc
            class_path: blueprints.VPC
            class_path: blueprints.Fake
        """
        with self.assertRaises(ConstructorError):
            parse(yaml_config)

    def test_raise_construct_error_on_duplicate_stack_name_dict(self):
        """Some mappings should never have a duplicate children.

        For example we treat `stacks` as a special mapping and disallow
        dupe children keys.

        """
        yaml_config = """
        namespace: prod
        stacks:
          my_vpc:
            class_path: blueprints.VPC1
          my_vpc:
            class_path: blueprints.VPC2
        """
        with self.assertRaises(ConstructorError):
            parse(yaml_config)

    def test_parse_invalid_inner_keys(self):
        """Test parse invalid innerkeys."""
        yaml_config = """
        namespace: prod
        stacks:
        - name: vpc
          class_path: blueprints.VPC
          garbage: yes
          variables:
            Foo: bar
        """

        with self.assertRaises(exceptions.InvalidConfig):
            parse(yaml_config)

    def test_stacker_to_runway_field_conversion(self):
        """Ensure the correct value is being presented by the config."""
        yaml_config = """
        namespace: test
        cfngin_bucket: ''
        stacker_bucket: test-bucket
        cfngin_bucket_region: us-east-1
        stacker_cache_dir: ./test/path
        """
        config = parse(yaml_config)
        # config.validate()

        assert config.cfngin_bucket == ''
        assert config.cfngin_bucket_region == 'us-east-1'
        assert config.cfngin_cache_dir == './test/path'


if __name__ == '__main__':
    unittest.main()



"""Retrieve a value from an environment variable.

The value is retrieved from a copy of the current environment variables
that is saved to the context object. These environment variables are
manipulated at runtime by Runway to fill in additional values such as
``DEPLOY_ENVIRONMENT`` and ``AWS_REGION`` to match the current execution.

.. note:: ``DEPLOY_ENVIRONMENT`` and ``AWS_REGION`` can only be resolved
          during the processing of a module. To ensure no error occurs
          when trying to resolve one of these in a :ref:`Deployment
          <runway-deployment>` definition, provide a default value.

If the Lookup is unable to find an environment variable matching the
provided query, the default value is returned or a ``ValueError`` is raised
if a default value was not provided.

.. rubric:: Arguments

This Lookup supports all :ref:`Common Lookup Arguments` but, the folling have
limited or no effect:

- region


.. rubric:: Example
.. code-block:: yaml

  deployment:
    - modules:
        - path: sampleapp.cfn
          parameters:
            creator: ${env USER}
      env_vars:
        ENVIRONMENT: ${env DEPLOY_ENVIRONMENT::default=default}

"""
# pylint: disable=arguments-differ
from typing import Any, TYPE_CHECKING  # pylint: disable=unused-import

from .base import LookupHandler

# python2 supported pylint sees this is cyclic even though its only for type checking
# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from ...context import Context  # noqa: F401 pylint: disable=unused-import

TYPE_NAME = "env"


class EnvLookup(LookupHandler):
    """Environment variable Lookup."""

    @classmethod
    def handle(cls, value, context, **_):
        # type: (str, 'Context', Any) -> Any
        """Retrieve an environment variable.

        The value is retrieved from a copy of the current environment variables
        that is saved to the context object. These environment variables
        are manipulated at runtime by Runway to fill in additional values
        such as ``DEPLOY_ENVIRONMENT`` and ``AWS_REGION`` to match the
        current execution.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        Raises:
            ValueError: Unable to find a value for the provided query and
                a default value was not provided.

        """
        query, args = cls.parse(value)

        result = context.env_vars.get(query, args.pop('default', ''))

        if result != '':  # allows for False bool and NoneType results
            return cls.format_results(result, **args)

        raise ValueError('"{}" does not exist in the environment'.format(value))



"""Retrieve a value from SSM Parameter Store.

If the Lookup is unable to find an SSM Parameter matching the provided query,
the default value is returned or ``ParameterNotFound`` is raised if a default
value is not provided.

Parameters of type ``SecureString`` are automatically decrypted.

Parameters of type ``StringList`` are returned as a list.


.. rubric:: Arguments

This Lookup supports all :ref:`Common Lookup Arguments`.


.. Example
.. code-block:: yaml

  deployment:
    - modules:
        - path: sampleapp.cfn
          parameters:
            secret_value: ${ssm /example/secret}
            conf_file: ${ssm /example/config/json::load=json, get=value}
            toggle: ${ssm toggle::load=yaml, get=val, transform=bool}
      env_vars:
        SOME_VARIABLE: ${ssm /example/param::region=us-east-1}
        DEFAULT_VARIABLE: ${ssm /example/default::default=default}

"""
# pylint: disable=arguments-differ
import logging
from typing import TYPE_CHECKING, Any, Union  # pylint: disable=unused-import

# using absolute for runway imports so stacker shim doesn't break when used from CFNgin
from runway.lookups.handlers.base import LookupHandler

# python2 supported pylint sees this is cyclic even though its only for type checking
# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from runway.context import Context as RunwayContext  # noqa: F401 pylint: disable=W
    from runway.cfngin.context import Context as CFNginContext  # noqa: F401 pylint: disable=W

LOGGER = logging.getLogger(__name__)
TYPE_NAME = 'ssm'


class SsmLookup(LookupHandler):
    """SSM Parameter Store Lookup."""

    @classmethod
    def handle(cls, value, context, **_):
        # type: (str, Union['CFNginContext', 'RunwayContext'], Any) -> Any
        """Retrieve a value from SSM Parameter Store.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        Raises:
            ParameterNotFound: Parameter not found in SSM and a default value
                was not provided.

        """
        query, args = cls.parse(value)

        session = context.get_session(region=args.get('region'))
        client = session.client('ssm')

        try:
            response = client.get_parameter(
                Name=query,
                WithDecryption=True
            )['Parameter']
            return cls.format_results(response['Value'].split(',')
                                      if response['Type'] == 'StringList'
                                      else response['Value'], **args)
        except client.exceptions.ParameterNotFound:
            if args.get('default'):
                args.pop('load', None)  # don't load a default value
                return cls.format_results(args.pop('default'), **args)
            raise



"""'Git' type Path Source."""
from __future__ import absolute_import
# pylint: disable=unused-import
from typing import List, Dict, Optional, Union  # noqa: F401

import tempfile
import shutil
import subprocess

import os
import sys
import logging

from .source import Source

LOGGER = logging.getLogger('runway')


class Git(Source):
    """Git Path Source.

    The Git path source can be tasked with cloning a remote repository
    and pointing to a specific module folder (or the root).

    """

    # Added for documentation purposes
    def __init__(self, uri='', location='', options=None, **kwargs):
        # type(Dict[str, Union[str, Dict[str, str]]]) -> Source
        """Git Path Source.

        Keyword Args:
            uri (str): The uniform resource identifier that targets the remote git
                repository
            location (string): The relative location to the root of the
                repository where the module resides. Leaving this as an empty
                string, ``/``, or ``./`` will have runway look in the root folder.
            options (Union(None, Dict[str, str])): A reference can be passed along via the
                options so that a specific version of the repository is cloned.
                **commit**, **tag**, **branch**  are all valid keys with
                respective output

        """
        self.uri = uri
        self.location = location
        self.options = options

        if not self.options:
            self.options = {}

        super(Git, self).__init__(**kwargs)

    def fetch(self):
        # type: () -> str
        """Retrieve the git repository from it's remote location."""
        from git import Repo

        ref = self.__determine_git_ref()  # type: str
        dir_name = '_'.join([self.sanitize_git_path(self.uri), ref])  # type: str
        cached_dir_path = os.path.join(self.cache_dir, dir_name)  # type: str
        cached_path = ''  # type: str

        if not os.path.isdir(cached_dir_path):
            tmp_dir = tempfile.mkdtemp()
            try:
                tmp_repo_path = os.path.join(tmp_dir, dir_name)  # type: str
                with Repo.clone_from(self.uri, tmp_repo_path) as repo:
                    repo.head.reference = ref  # type: str
                    repo.head.reset(index=True, working_tree=True)
                shutil.move(tmp_repo_path, self.cache_dir)
                cached_path = os.path.join(self.cache_dir, dir_name)  # type: str
            finally:
                shutil.rmtree(tmp_dir)
        else:
            cached_path = cached_dir_path  # type: str

        return os.path.join(cached_path, self.location)

    def __git_ls_remote(self, ref):
        # type: (str) -> str
        """List remote repositories based on uri and ref received.

        Keyword Args:
            ref (str): The git reference value

        """
        LOGGER.debug("Invoking git to retrieve commit id for repo %s...", self.uri)
        lsremote_output = subprocess.check_output(['git', 'ls-remote', self.uri, ref])
        # pylint: disable=unsupported-membership-test
        if b"\t" in lsremote_output:
            commit_id = lsremote_output.split(b"\t")[0]  # type List[str]
            LOGGER.debug("Matching commit id found: %s", commit_id)
            return commit_id
        raise ValueError("Ref \"%s\" not found for repo %s." % (ref, self.uri))

    def __determine_git_ls_remote_ref(self):
        # type: () -> str
        """Determine remote ref, defaulting to HEAD unless a branch is found."""
        ref = "HEAD"

        if self.options.get('branch'):
            ref = "refs/heads/%s" % self.options.get('branch')  # type: str

        return ref

    def __determine_git_ref(self):
        # type: () -> str
        """Determine the git reference code."""
        ref_config_keys = 0   # type: int

        for i in ['commit', 'tag', 'branch']:
            if self.options.get(i):
                ref_config_keys += 1
        if ref_config_keys > 1:
            raise ImportError("Fetching remote git sources failed: "
                              "conflicting revisions (e.g. 'commit', 'tag', "
                              "'branch') specified for a package source")

        if self.options.get('commit'):
            ref = self.options.get('commit')  # type: str
        elif self.options.get('tag'):
            ref = self.options.get('tag')  # type: str
        else:
            ref = self.__git_ls_remote(self.__determine_git_ls_remote_ref())  # ty pe: str
        if sys.version_info[0] > 2 and isinstance(ref, bytes):
            return ref.decode()
        return ref

    @classmethod
    def sanitize_git_path(cls, path):
        # type(str) -> str
        """Sanitize the git path for folder/file assignment.

        Keyword Args:
            path (str): The path string to be sanitized

        """
        dir_name = path  # type: str
        split = path.split('//')  # type: List[str]
        domain = split[len(split) - 1]  # type: str

        if domain.endswith('.git'):
            dir_name = domain[:-4]

        return cls.sanitize_directory_path(dir_name)



"""Abstraction for the module 'type' value in a a Runway configuration."""
# pylint: disable=unused-import
from typing import Dict, List, Optional, Tuple, Union  # noqa: F401

import glob
import logging
import os
import sys

from .util import (load_object_from_string)

LOGGER = logging.getLogger('runway')


# noqa pylint: disable=too-few-public-methods
class RunwayModuleType(object):
    """Runway configuration ``type`` settings object.

    The ``type`` property of a Runway configuration can be
    used to explicitly specify what module type you are
    intending to deploy.

    Runway determines the type of module you are trying to
    deploy in 3 different ways. First, it will check for the
    ``type`` property as described here, next it will look
    for a suffix as described in :ref:`Module Definition<mod-definition>`,
    and finally it will attempt to autodetect your module
    type by scanning the files of the project. If none of
    those settings produces a valid result an error will
    occur. The following are valid explicit types:

    +--------------------+-----------------------------------------------+
    | Type               | IaC Tool/Framework                            |
    +====================+===============================================+
    | ``cdk``            | `AWS CDK`_                                    |
    +--------------------+-----------------------------------------------+
    | ``cloudformation`` | `CloudFormation`_                             |
    +--------------------+-----------------------------------------------+
    | ``serverless``     | `Serverless Framework`_                       |
    +--------------------+-----------------------------------------------+
    | ``terraform``      | `Terraform`_                                  |
    +--------------------+-----------------------------------------------+
    | ``kubernetes``     | `Kubernetes`_                                 |
    +--------------------+-----------------------------------------------+
    | ``static``         | :ref:`Static Site<mod-staticsite>`            |
    +--------------------+-----------------------------------------------+

    Even when specifying a module ``type`` the module structure
    needs to be conducive with that type of project. If the files contained
    within don't match the type then an error will occur.

    """

    EXTENSION_MAP = {
        'sls': 'runway.module.serverless.Serverless',
        'tf': 'runway.module.terraform.Terraform',
        'cdk': 'runway.module.cdk.CloudDevelopmentKit',
        'k8s': 'runway.module.k8s.K8s',
        'cfn': 'runway.module.cloudformation.CloudFormation',
        'web': 'runway.module.staticsite.StaticSite'
    }

    TYPE_MAP = {
        'serverless': EXTENSION_MAP.get('sls'),
        'terraform': EXTENSION_MAP.get('tf'),
        'cdk': EXTENSION_MAP.get('cdk'),
        'kubernetes': EXTENSION_MAP.get('k8s'),
        'cloudformation': EXTENSION_MAP.get('cfn'),
        'static': EXTENSION_MAP.get('web'),
    }

    def __init__(self, path, class_path=None, type_str=None):
        # type: (str, Optional[str], Optional[str]) -> RunwayModuleType
        """Initialization of the Module Type Configuration.  # noqa

        Keyword Args:
            path (str): The required path to the module
            class_path (Optional[str]): A supplied class_path to override
                the autodetected one.
            type_str (Optional[str]): An explicit type to assign to
                the RunwayModuleType
        """
        self.path = path
        self.class_path = class_path
        self.type_str = type_str
        self.module_class = self._determine_module_class()

    def _determine_module_class(self):
        """Determine type of module and return deployment module class.

        Returns:
            object: The specified module class

        """
        if not self.class_path:
            self._set_class_path_based_on_extension()

        if not self.class_path and self.type_str:
            self.class_path = self.TYPE_MAP.get(self.type_str, None)

        if not self.class_path:
            self._set_class_path_based_on_autodetection()

        if not self.class_path:
            LOGGER.error('No module class found for %s', os.path.basename(self.path))
            sys.exit(1)

        return load_object_from_string(self.class_path)

    def _set_class_path_based_on_extension(self):
        # type() -> void
        """Based on the directory suffix set the class_path."""
        basename = os.path.basename(self.path)
        basename_split = basename.split('.')
        extension = basename_split[len(basename_split) - 1]
        self.class_path = self.EXTENSION_MAP.get(extension, None)

    def _set_class_path_based_on_autodetection(self):
        # type() -> void
        """Based on the files detected in the base path set the class_path."""
        if (self._is_file('serverless.yml') or self._is_file('serverless.js')) \
                and self._is_file('package.json'):
            self.class_path = self.TYPE_MAP.get('serverless', None)
        elif self._has_glob('*.tf'):
            self.class_path = self.TYPE_MAP.get('terraform', None)
        elif self._is_file('cdk.json') and self._is_file('package.json'):
            self.class_path = self.TYPE_MAP.get('cdk', None)
        elif self._is_dir('overlays') and self._find_kustomize_files():
            self.class_path = self.TYPE_MAP.get('kubernetes', None)
        elif self._has_glob('*.env') \
            or self._has_glob('*.yaml') \
                or self._has_glob('*.yml'):
            self.class_path = self.TYPE_MAP.get('cloudformation', None)

    def _is_file(self, file_name):
        # type(str) -> boolean
        """Verify if specified filename is a file.

        Keyword Args:
            file_name (str): The filename relative to the initialized path

        Returns:
            boolean: Whether the file_name passed in is a file

        """
        return os.path.isfile(os.path.join(self.path, file_name))

    def _is_dir(self, dir_name):
        # type: (str) -> bool
        """Verify if specified dir_name is a directory.

        Keyword Args:
            dir_name (str): The directory relative to the initialized path

        Returns:
            boolean: Whether the passed in dir_name is a directory

        """
        return os.path.isdir(os.path.join(self.path, dir_name))

    def _has_glob(self, glb):
        # type: (str)-> bool
        """Verify if a glob of files exist.

        Keyword Args:
            glb (str): The glob pattern relative to the initialized path

        Returns:
            boolean: Whether the passed in glob of files exist

        """
        return glob.glob(os.path.join(self.path, glb))

    def _find_kustomize_files(self):
        # type: ()-> bool
        """Return true if kustomize yaml file found.

        Returns:
            boolean: Whether the kustomize yaml exist

        """
        for _root, _dirnames, filenames in os.walk(self.path):
            for filename in filenames:
                if filename == 'kustomization.yaml':
                    return True
        return False
